# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import sys
import logging

from ..exceptions import ConfigurationInvalid, UpdateError, CannotUpdateOffline

from octoprint.util.commandline import CommandlineCaller, CommandlineError


def _get_caller(log_cb=None):
	def _log_call(*lines):
		_log(lines, prefix=" ", stream="call")

	def _log_stdout(*lines):
		_log(lines, prefix=">", stream="stdout")

	def _log_stderr(*lines):
		_log(lines, prefix="!", stream="stderr")

	def _log(lines, prefix=None, stream=None):
		if log_cb is None:
			return
		log_cb(lines, prefix=prefix, stream=stream)

	caller = CommandlineCaller()
	if log_cb is not None:
		caller.on_log_call = _log_call
		caller.on_log_stdout = _log_stdout
		caller.on_log_stderr = _log_stderr
	return caller


def can_perform_update(target, check, online=True):
	import os
	script_configured = bool("update_script" in check and check["update_script"])

	folder = None
	if "update_folder" in check:
		folder = check["update_folder"]
	elif "checkout_folder" in check:
		folder = check["checkout_folder"]
	folder_configured = bool(folder and os.path.isdir(folder))

	return script_configured and folder_configured and (online or check.get("offline", False))


def perform_update(target, check, target_version, log_cb=None, online=True):
	logger = logging.getLogger("octoprint.plugins.softwareupdate.updaters.update_script")

	if not online and not check("offline", False):
		raise CannotUpdateOffline()

	if not can_perform_update(target, check):
		raise ConfigurationInvalid("checkout_folder and update_folder are missing for update target %s, one is needed" % target)

	update_script = check["update_script"]
	update_branch = check.get("update_branch", "")
	force_exact_version = check.get("force_exact_version", False)
	folder = check.get("update_folder", check.get("checkout_folder")) # either should be set, tested above
	pre_update_script = check.get("pre_update_script", None)
	post_update_script = check.get("post_update_script", None)

	caller = _get_caller(log_cb=log_cb)

	### pre update

	if pre_update_script is not None:
		logger.debug("Target: %s, running pre-update script: %s" % (target, pre_update_script))
		try:
			caller.checked_call(pre_update_script, cwd=folder)
		except CommandlineError as e:
			logger.exception("Target: %s, error while executing pre update script, got returncode %r" % (target, e.returncode))

	### update

	try:
		update_command = update_script.format(python=sys.executable,
		                                      folder=folder,
		                                      target=target_version,
		                                      branch=update_branch,
		                                      force="true" if force_exact_version else "false")

		logger.debug("Target %s, running update script: %s" % (target, update_command))

		caller.checked_call(update_command, cwd=folder)
	except CommandlineError as e:
		logger.exception("Target: %s, error while executing update script, got returncode %r" % (target, e.returncode))
		raise UpdateError("Error while executing update script for %s", (e.stdout, e.stderr))

	### post update

	if post_update_script is not None:
		logger.debug("Target: %s, running post-update script %s..." % (target, post_update_script))
		try:
			caller.checked_call(post_update_script, cwd=folder)
		except CommandlineError as e:
			logger.exception("Target: %s, error while executing post update script, got returncode %r" % (target, e.returncode))

	return "ok"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, make_response, jsonify

from octoprint.server import printer, NO_CONTENT, current_user
from octoprint.server.util.flask import no_firstrun_access, get_json_command_from_request
from octoprint.server.api import api
from octoprint.access.permissions import Permissions

@api.route("/job", methods=["POST"])
@no_firstrun_access
def controlJob():
	if not printer.is_operational():
		return make_response("Printer is not operational", 409)

	valid_commands = {
		"start": [],
		"restart": [],
		"pause": [],
		"cancel": []
	}

	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	activePrintjob = printer.is_printing() or printer.is_paused()

	tags = {"source:api", "api:job"}

	user = current_user.get_name()

	with Permissions.PRINT.require(403):
		if command == "start":
			if activePrintjob:
				return make_response("Printer already has an active print job, did you mean 'restart'?", 409)
			printer.start_print(tags=tags, user=user)
		elif command == "restart":
			if not printer.is_paused():
				return make_response("Printer does not have an active print job or is not paused", 409)
			printer.start_print(tags=tags, user=user)
		elif command == "pause":
			if not activePrintjob:
				return make_response("Printer is neither printing nor paused, 'pause' command cannot be performed", 409)
			action = data.get("action", "toggle")
			if action == "toggle":
				printer.toggle_pause_print(tags=tags, user=user)
			elif action == "pause":
				printer.pause_print(tags=tags, user=user)
			elif action == "resume":
				printer.resume_print(tags=tags, user=user)
			else:
				return make_response("Unknown action '{}', allowed values for action parameter are 'pause', 'resume' and 'toggle'".format(action), 400)
		elif command == "cancel":
			if not activePrintjob:
				return make_response("Printer is neither printing nor paused, 'cancel' command cannot be performed", 409)
			printer.cancel_print(tags=tags, user=user)
	return NO_CONTENT


@api.route("/job", methods=["GET"])
@Permissions.STATUS.require(403)
def jobState():
	currentData = printer.get_current_data()
	return jsonify({
		"job": currentData["job"],
		"progress": currentData["progress"],
		"state": currentData["state"]["text"]
	})



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin
import logging

from octoprint.util import get_fully_qualified_classname as fqcn
from octoprint.util.version import get_octoprint_version_string, is_released_octoprint_version

from flask import jsonify
from flask_babel import gettext

SENTRY_URL_SERVER = "https://2e668ba966024cfc962ba97eef89dc15@sentry.io/1373987"
SENTRY_URL_COREUI = "https://dccc2e2f3c6b4adda72f376bbb7fe561@sentry.io/1374096"

SETTINGS_DEFAULTS = dict(enabled=False,
                         enabled_unreleased=False,
                         unique_id=None,
                         url_server=SENTRY_URL_SERVER,
                         url_coreui=SENTRY_URL_COREUI)

import serial
import requests.exceptions
import errno
import octoprint.util.avr_isp.ispBase
import tornado.websocket

IGNORED_EXCEPTIONS = [
	# serial exceptions in octoprint.util.comm
	(serial.SerialException, lambda exc, logger, plugin, cb: logger == "octoprint.util.comm"),

	# isp errors during port auto detection in octoprint.util.comm
	(octoprint.util.avr_isp.ispBase.IspError, lambda exc, logger, plugin, cb: logger == "octoprint.util.comm"),

	# IOErrors of any kind due to a full file system
	(IOError, lambda exc, logger, plugin, cb: getattr(exc, "errno") and exc.errno in (getattr(errno, "ENOSPC"),)),

	# RequestExceptions of any kind
	requests.exceptions.RequestException,

	# Tornado WebSocketErrors of any kind
	tornado.websocket.WebSocketError,

	# Anything triggered by or in third party plugin Astroprint
	(Exception, lambda exc, logger, plugin, cb: logger.startswith("octoprint.plugins.astroprint")
	                                            or plugin == "astroprint"
	                                            or cb.startswith("octoprint_astroprint."))
]

try:
	# noinspection PyUnresolvedReferences
	from octoprint.plugins.backup import InsufficientSpace

	# if the backup plugin is enabled, ignore InsufficientSpace errors from it as well
	IGNORED_EXCEPTIONS.append(InsufficientSpace)

	del InsufficientSpace
except ImportError:
	pass


class ErrorTrackingPlugin(octoprint.plugin.SettingsPlugin,
                          octoprint.plugin.AssetPlugin,
                          octoprint.plugin.TemplatePlugin,
                          octoprint.plugin.SimpleApiPlugin):

	def get_template_configs(self):
		return [
			dict(type="settings",
			     name=gettext("Error Tracking"),
			     template="errortracking_settings.jinja2",
			     custom_bindings=False),
			dict(type="generic", template="errortracking_javascripts.jinja2")
		]

	def get_template_vars(self):
		enabled = self._settings.get_boolean(["enabled"])
		enabled_unreleased = self._settings.get_boolean(["enabled_unreleased"])

		return dict(enabled=_is_enabled(enabled, enabled_unreleased),
		            unique_id=self._settings.get(["unique_id"]),
		            url_coreui=self._settings.get(["url_coreui"]))

	def get_assets(self):
		return dict(js=["js/sentry.min.js", "js/errortracking.js"])

	def get_settings_defaults(self):
		return SETTINGS_DEFAULTS

	def on_settings_save(self, data):
		old_enabled = _is_enabled(self._settings.get_boolean(["enabled"]),
		                          self._settings.get_boolean(["enabled_unreleased"]))

		octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

		enabled = _is_enabled(self._settings.get_boolean(["enabled"]),
		                      self._settings.get_boolean(["enabled_unreleased"]))

		if old_enabled != enabled:
			_enable_errortracking()

	def on_api_get(self, request):
		return jsonify(**self.get_template_vars())


_enabled = False
def _enable_errortracking():
	# this is a bit hackish, but we want to enable error tracking as early in the platform lifecycle as possible
	# and hence can't wait until our implementation is initialized and injected with settings

	from octoprint.settings import settings
	global _enabled

	if _enabled:
		return

	version = get_octoprint_version_string()

	s = settings()
	plugin_defaults = dict(plugins=dict(errortracking=SETTINGS_DEFAULTS))

	enabled = s.getBoolean(["plugins", "errortracking", "enabled"], defaults=plugin_defaults)
	enabled_unreleased = s.getBoolean(["plugins", "errortracking", "enabled_unreleased"], defaults=plugin_defaults)
	url_server = s.get(["plugins", "errortracking", "url_server"], defaults=plugin_defaults)
	unique_id = s.get(["plugins", "errortracking", "unique_id"], defaults=plugin_defaults)
	if unique_id is None:
		import uuid
		unique_id = str(uuid.uuid4())
		s.set(["plugins", "errortracking", "unique_id"], unique_id, defaults=plugin_defaults)
		s.save()

	if _is_enabled(enabled, enabled_unreleased):
		import sentry_sdk
		from octoprint.plugin import plugin_manager

		def _before_send(event, hint):
			if not "exc_info" in hint:
				# we only want exceptions
				return None

			handled = True
			logger = event.get("logger", "")
			plugin = event.get("extra", dict()).get("plugin", None)
			callback = event.get("extra", dict()).get("callback", None)

			for ignore in IGNORED_EXCEPTIONS:
				if isinstance(ignore, tuple):
					ignored_exc, matcher = ignore
				else:
					ignored_exc = ignore
					matcher = lambda *args: True

				exc = hint["exc_info"][1]
				if isinstance(exc, ignored_exc) and matcher(exc, logger, plugin, callback):
					# exception ignored for logger, plugin and/or callback
					return None

				elif isinstance(ignore, type):
					if isinstance(hint["exc_info"][1], ignore):
						# exception ignored
						return None

			if event.get("exception") and event["exception"].get("values"):
				handled = not any(map(lambda x: x.get("mechanism") and x["mechanism"].get("handled", True) == False,
				                      event["exception"]["values"]))

			if handled:
				# error is handled, restrict further based on logger
				if logger != "" and not (logger.startswith("octoprint.") or logger.startswith("tornado.")):
					# we only want errors logged by loggers octoprint.* or tornado.*
					return None

				if logger.startswith("octoprint.plugins."):
					plugin_id = logger.split(".")[2]
					plugin_info = plugin_manager().get_plugin_info(plugin_id)
					if plugin_info is None or not plugin_info.bundled:
						# we only want our active bundled plugins
						return None

				if plugin is not None:
					plugin_info = plugin_manager().get_plugin_info(plugin)
					if plugin_info is None or not plugin_info.bundled:
						# we only want our active bundled plugins
						return None

			return event

		sentry_sdk.init(url_server,
		                release=version,
		                before_send=_before_send)

		with sentry_sdk.configure_scope() as scope:
			scope.user = dict(id=unique_id)

		logging.getLogger("octoprint.plugins.errortracking").info("Initialized error tracking")
		_enabled = True


def _is_enabled(enabled, enabled_unreleased):
	return enabled and (enabled_unreleased or is_released_octoprint_version())


def __plugin_enable__():
	_enable_errortracking()


__plugin_name__ = "Error Tracking"
__plugin_author__ = "Gina Häußge"
__plugin_implementation__ = ErrorTrackingPlugin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import logging

from flask import Blueprint, request, jsonify, abort, current_app, session, make_response, g
from flask_login import login_user, logout_user, current_user
from werkzeug.exceptions import HTTPException
from octoprint.vendor.flask_principal import Identity, identity_changed, AnonymousIdentity

import octoprint.access.users
import octoprint.util.net as util_net
import octoprint.server
import octoprint.plugin
from octoprint.server import NO_CONTENT
from octoprint.settings import settings as s, valid_boolean_trues
from octoprint.server.util import noCachingExceptGetResponseHandler, loginFromApiKeyRequestHandler, loginFromAuthorizationHeaderRequestHandler, corsRequestHandler, corsResponseHandler
from octoprint.server.util.flask import no_firstrun_access, get_json_command_from_request, passive_login, get_remote_address
from octoprint.access.permissions import Permissions
from octoprint.events import eventManager, Events


#~~ init api blueprint, including sub modules

api = Blueprint("api", __name__)

from . import printer as api_printer
from . import job as api_job
from . import connection as api_connection
from . import files as api_files
from . import settings as api_settings
from . import timelapse as api_timelapse
from . import access as api_access
from . import users as api_users
from . import slicing as api_slicing
from . import printer_profiles as api_printer_profiles
from . import languages as api_languages
from . import system as api_system


VERSION = "0.1"

api.after_request(noCachingExceptGetResponseHandler)

api.before_request(corsRequestHandler)
api.before_request(loginFromAuthorizationHeaderRequestHandler)
api.before_request(loginFromApiKeyRequestHandler)
api.after_request(corsResponseHandler)

#~~ data from plugins

@api.route("/plugin/<string:name>", methods=["GET"])
def pluginData(name):
	api_plugins = octoprint.plugin.plugin_manager().get_filtered_implementations(lambda p: p._identifier == name, octoprint.plugin.SimpleApiPlugin)
	if not api_plugins:
		return make_response("Not found", 404)

	if len(api_plugins) > 1:
		return make_response("More than one api provider registered for {name}, can't proceed".format(name=name), 500)

	try:
		api_plugin = api_plugins[0]
		if api_plugin.is_api_adminonly() and not current_user.is_admin:
			return make_response("Forbidden", 403)

		response = api_plugin.on_api_get(request)

		if response is not None:
			return response
		return NO_CONTENT
	except HTTPException:
		raise
	except Exception:
		logging.getLogger(__name__).exception("Error calling SimpleApiPlugin {}".format(name),
		                                      extra=dict(plugin=name))
		return abort(500)

#~~ commands for plugins

@api.route("/plugin/<string:name>", methods=["POST"])
@no_firstrun_access
def pluginCommand(name):
	api_plugins = octoprint.plugin.plugin_manager().get_filtered_implementations(lambda p: p._identifier == name, octoprint.plugin.SimpleApiPlugin)

	if not api_plugins:
		return make_response("Not found", 404)

	if len(api_plugins) > 1:
		return make_response("More than one api provider registered for {name}, can't proceed".format(name=name), 500)

	api_plugin = api_plugins[0]
	try:
		valid_commands = api_plugin.get_api_commands()
		if valid_commands is None:
			return make_response("Method not allowed", 405)

		if api_plugin.is_api_adminonly() and not Permissions.ADMIN.can():
			return make_response("Forbidden", 403)

		command, data, response = get_json_command_from_request(request, valid_commands)
		if response is not None:
			return response

		response = api_plugin.on_api_command(command, data)
		if response is not None:
			return response
		return NO_CONTENT
	except HTTPException:
		raise
	except Exception:
		logging.getLogger(__name__).exception("Error while executing SimpleApiPlugin {}".format(name),
		                                      extra=dict(plugin=name))
		return abort(500)

#~~ first run setup

@api.route("/setup/wizard", methods=["GET"])
def wizardState():
	if not s().getBoolean(["server", "firstRun"]) and not Permissions.ADMIN.can():
		abort(403)

	seen_wizards = s().get(["server", "seenWizards"])

	result = dict()
	wizard_plugins = octoprint.server.pluginManager.get_implementations(octoprint.plugin.WizardPlugin)
	for implementation in wizard_plugins:
		name = implementation._identifier
		try:
			required = implementation.is_wizard_required()
			details = implementation.get_wizard_details()
			version = implementation.get_wizard_version()
			ignored = octoprint.plugin.WizardPlugin.is_wizard_ignored(seen_wizards, implementation)
		except Exception:
			logging.getLogger(__name__).exception("There was an error fetching wizard "
			                                      "details for {}, ignoring".format(name),
			                                      extra=dict(plugin=name))
		else:
			result[name] = dict(required=required, details=details, version=version, ignored=ignored)

	return jsonify(result)


@api.route("/setup/wizard", methods=["POST"])
def wizardFinish():
	if not s().getBoolean(["server", "firstRun"]) and not Permissions.ADMIN.can():
		abort(403)

	data = dict()
	try:
		data = request.get_json()
	except Exception:
		abort(400)

	if data is None:
		abort(400)

	if not "handled" in data:
		abort(400)
	handled = data["handled"]

	if s().getBoolean(["server", "firstRun"]):
		s().setBoolean(["server", "firstRun"], False)

	seen_wizards = dict(s().get(["server", "seenWizards"]))

	wizard_plugins = octoprint.server.pluginManager.get_implementations(octoprint.plugin.WizardPlugin)
	for implementation in wizard_plugins:
		name = implementation._identifier
		try:
			implementation.on_wizard_finish(name in handled)
			if name in handled:
				seen_wizards[name] = implementation.get_wizard_version()
		except Exception:
			logging.getLogger(__name__).exception("There was an error finishing the "
			                                      "wizard for {}, ignoring".format(name),
			                                      extra=dict(plugin=name))

	s().set(["server", "seenWizards"], seen_wizards)
	s().save()

	return NO_CONTENT


#~~ system state


@api.route("/state", methods=["GET"])
@no_firstrun_access
def apiPrinterState():
	return make_response(("/api/state has been deprecated, use /api/printer instead", 405, []))


@api.route("/version", methods=["GET"])
def apiVersion():
	return jsonify({
		"server": octoprint.server.VERSION,
		"api": VERSION,
		"text": "OctoPrint {}".format(octoprint.server.DISPLAY_VERSION)
	})


#~~ Login/user handling


@api.route("/login", methods=["POST"])
def login():
	data = request.get_json()
	if not data:
		data = request.values

	if octoprint.server.userManager.enabled and "user" in data and "pass" in data:
		username = data["user"]
		password = data["pass"]

		if "remember" in data and data["remember"] in valid_boolean_trues:
			remember = True
		else:
			remember = False

		if "usersession.id" in session:
			_logout(current_user)

		user = octoprint.server.userManager.find_user(username)
		if user is not None:
			if octoprint.server.userManager.check_password(username, password):
				if not user.is_active:
					return make_response(("Your account is deactivated", 403, []))

				if octoprint.server.userManager.enabled:
					user = octoprint.server.userManager.login_user(user)
					session["usersession.id"] = user.session
					g.user = user
				login_user(user, remember=remember)
				identity_changed.send(current_app._get_current_object(), identity=Identity(user.get_id()))

				remote_addr = get_remote_address(request)
				logging.getLogger(__name__).info("Actively logging in user {} from {}".format(user.get_id(), remote_addr))

				response = user.as_dict()
				response["_is_external_client"] = s().getBoolean(["server", "ipCheck", "enabled"]) \
				                                  and not util_net.is_lan_address(remote_addr,
				                                                                  additional_private=s().get(["server", "ipCheck", "trustedSubnets"]))

				r = make_response(jsonify(response))
				r.delete_cookie("active_logout")

				eventManager().fire(Events.USER_LOGGED_IN, payload=dict(username=user.get_id()))

				return r

		return make_response(("User unknown or password incorrect", 403, []))

	elif "passive" in data:
		return passive_login()

	return make_response("Neither user and pass attributes nor passive flag present", 400)


@api.route("/logout", methods=["POST"])
def logout():
	username = None
	if current_user:
		username = current_user.get_id()

	# logout from user manager...
	_logout(current_user)

	# ... and from flask login (and principal)
	logout_user()

	# ... and send an active logout session cookie
	r = make_response(jsonify(octoprint.server.userManager.anonymous_user_factory()))
	r.set_cookie("active_logout", "true")

	if username:
		eventManager().fire(Events.USER_LOGGED_OUT, payload=dict(username=username))

	return r


def _logout(user):
	if "usersession.id" in session:
		del session["usersession.id"]
	octoprint.server.userManager.logout_user(user)


@api.route("/currentuser", methods=["GET"])
def get_current_user():
	return jsonify(name=current_user.get_name(),
	               permissions=[permission.key for permission in current_user.effective_permissions],
	               groups=[group.key for group in current_user.groups])


#~~ Test utils


@api.route("/util/test", methods=["POST"])
@no_firstrun_access
@Permissions.ADMIN.require(403)
def utilTest():
	valid_commands = dict(
		path=["path"],
		url=["url"],
		server=["host", "port"]
	)

	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	if command == "path":
		return _test_path(data)
	elif command == "url":
		return _test_url(data)
	elif command == "server":
		return _test_server(data)

def _test_path(data):
	import os
	from octoprint.util.paths import normalize

	path = normalize(data["path"], real=False)
	if not path:
		return jsonify(path=path, exists=False, typeok=False, broken_symlink=False, access=False, result=False)

	unreal_path = path
	path = os.path.realpath(path)

	check_type = None
	check_access = []

	if "check_type" in data and data["check_type"] in ("file", "dir"):
		check_type = data["check_type"]

	if "check_access" in data:
		request_check_access = data["check_access"]
		if not isinstance(request_check_access, list):
			request_check_access = list(request_check_access)

		check_access = [check for check in request_check_access if check in ("r", "w", "x")]

	allow_create_dir = data.get("allow_create_dir", False) and check_type == "dir"
	check_writable_dir = data.get("check_writable_dir", False) and check_type == "dir"
	if check_writable_dir and "w" not in check_access:
		check_access.append("w")

	# check if path exists
	exists = os.path.exists(path)
	if not exists:
		if os.path.islink(unreal_path):
			# broken symlink, see #2644
			logging.getLogger(__name__).error("{} is a broken symlink pointing at non existing {}".format(unreal_path, path))
			return jsonify(path=unreal_path, exists=False, typeok=False, broken_symlink=True, access=False, result=False)

		elif check_type == "dir" and allow_create_dir:
			try:
				os.makedirs(path)
			except Exception:
				logging.getLogger(__name__).exception("Error while trying to create {}".format(path))
				return jsonify(path=path, exists=False, typeok=False, broken_symlink=False, access=False, result=False)
			else:
				exists = True

	# check path type
	type_mapping = dict(file=os.path.isfile, dir=os.path.isdir)
	if check_type:
		typeok = type_mapping[check_type](path)
	else:
		typeok = exists

	# check if path allows requested access
	access_mapping = dict(r=os.R_OK, w=os.W_OK, x=os.X_OK)
	if check_access:
		mode = 0
		for a in map(lambda x: access_mapping[x], check_access):
			mode |= a
		access = os.access(path, mode)
	else:
		access = exists

	if check_writable_dir and check_type == "dir":
		try:
			test_path = os.path.join(path, ".testballoon.txt")
			with io.open(test_path, 'wb') as f:
				f.write(b"Test")
			os.remove(test_path)
		except Exception:
			logging.getLogger(__name__).exception("Error while testing if {} is really writable".format(path))
			return jsonify(path=path, exists=exists, typeok=typeok, broken_symlink=False, access=False, result=False)

	return jsonify(path=path, exists=exists, typeok=typeok, broken_symlink=False, access=access, result=exists and typeok and access)


def _test_url(data):
	import requests
	from octoprint import util as util

	class StatusCodeRange(object):
		def __init__(self, start=None, end=None):
			self.start = start
			self.end = end

		def __contains__(self, item):
			if not isinstance(item, int):
				return False
			if self.start and self.end:
				return self.start <= item < self.end
			elif self.start:
				return self.start <= item
			elif self.end:
				return item < self.end
			else:
				return False

		def as_dict(self):
			return dict(
				start=self.start,
				end=self.end
			)

	status_ranges = dict(
		informational=StatusCodeRange(start=100, end=200),
		success=StatusCodeRange(start=200, end=300),
		redirection=StatusCodeRange(start=300, end=400),
		client_error=StatusCodeRange(start=400, end=500),
		server_error=StatusCodeRange(start=500, end=600),
		normal=StatusCodeRange(end=400),
		error=StatusCodeRange(start=400, end=600),
		any=StatusCodeRange(start=100),
		timeout=StatusCodeRange(start=0, end=1)
	)

	url = data["url"]
	method = data.get("method", "HEAD")
	timeout = 3.0
	valid_ssl = True
	check_status = [status_ranges["normal"]]
	content_type_whitelist = None
	content_type_blacklist = None

	if "timeout" in data:
		try:
			timeout = float(data["timeout"])
		except Exception:
			return make_response("{!r} is not a valid value for timeout (must be int or float)".format(data["timeout"]),
			                     400)

	if "validSsl" in data:
		valid_ssl = data["validSsl"] in valid_boolean_trues

	if "status" in data:
		request_status = data["status"]
		if not isinstance(request_status, list):
			request_status = [request_status]

		check_status = []
		for rs in request_status:
			if isinstance(rs, int):
				check_status.append([rs])
			else:
				if rs in status_ranges:
					check_status.append(status_ranges[rs])
				else:
					code = requests.codes[rs]
					if code is not None:
						check_status.append([code])

	if "content_type_whitelist" in data:
		if not isinstance(data["content_type_whitelist"], (list, tuple)):
			return make_response("content_type_whitelist must be a list of mime types")
		content_type_whitelist = list(map(util.parse_mime_type, data["content_type_whitelist"]))
	if "content_type_blacklist" in data:
		if not isinstance(data["content_type_whitelist"], (list, tuple)):
			return make_response("content_type_blacklist must be a list of mime types")
		content_type_blacklist = list(map(util.parse_mime_type, data["content_type_blacklist"]))

	response_result = None
	outcome = True
	status = 0
	try:
		with requests.request(method=method, url=url, timeout=timeout, verify=valid_ssl, stream=True) as response:
			status = response.status_code
			outcome = outcome and any(map(lambda x: status in x, check_status))
			content_type = response.headers.get("content-type")

			response_result = dict(headers=dict(response.headers),
			                       content_type=content_type)

			if not content_type and data.get("content_type_guess") in valid_boolean_trues:
				content = response.content
				content_type = util.guess_mime_type(bytearray(content))

			if not content_type:
				content_type = "application/octet-stream"

			response_result = dict(assumed_content_type=content_type)

			parsed_content_type = util.parse_mime_type(content_type)

			in_whitelist = content_type_whitelist is None or any(
				map(lambda x: util.mime_type_matches(parsed_content_type, x), content_type_whitelist))
			in_blacklist = content_type_blacklist is not None and any(
				map(lambda x: util.mime_type_matches(parsed_content_type, x), content_type_blacklist))

			if not in_whitelist or in_blacklist:
				# we don't support this content type
				response.close()
				outcome = False

			elif "response" in data and (
					data["response"] in valid_boolean_trues or data["response"] in ("json", "bytes")):
				if data["response"] == "json":
					content = response.json()

				else:
					import base64
					content = base64.standard_b64encode(response.content)

				response_result["content"] = content
	except Exception:
		logging.getLogger(__name__).exception("Error while running a test {} request on {}".format(method, url))
		outcome = False

	result = dict(
		url=url,
		status=status,
		result=outcome
	)
	if response_result:
		result["response"] = response_result

	return jsonify(**result)

def _test_server(data):
	host = data["host"]
	try:
		port = int(data["port"])
	except Exception:
		return make_response("{!r} is not a valid value for port (must be int)".format(data["port"]), 400)

	timeout = 3.05
	if "timeout" in data:
		try:
			timeout = float(data["timeout"])
		except Exception:
			return make_response("{!r} is not a valid value for timeout (must be int or float)".format(data["timeout"]),
			                     400)

	protocol = data.get("protocol", "tcp")
	if protocol not in ("tcp", "udp"):
		return make_response("{!r} is not a valid value for protocol, must be tcp or udp".format(protocol), 400)

	from octoprint.util import server_reachable
	reachable = server_reachable(host, port, timeout=timeout, proto=protocol)

	result = dict(host=host,
	              port=port,
	              protocol=protocol,
	              result=reachable)

	return jsonify(**result)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import copy

from flask import jsonify, make_response, request, url_for
from werkzeug.exceptions import BadRequest

from past.builtins import basestring

from octoprint.server.api import api, NO_CONTENT, valid_boolean_trues
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking
from octoprint.util import dict_merge

from octoprint.server import printerProfileManager
from octoprint.printer.profile import InvalidProfileError, CouldNotOverwriteError
from octoprint.access.permissions import Permissions

def _lastmodified():
	return printerProfileManager.last_modified

def _etag(lm=None):
	if lm is None:
		lm = _lastmodified()

	import hashlib
	hash = hashlib.sha1()
	def hash_update(value):
		value = value.encode('utf-8')
		hash.update(value)
	hash_update(str(lm))
	hash_update(repr(printerProfileManager.get_default()))
	hash_update(repr(printerProfileManager.get_current()))
	return hash.hexdigest()


@api.route("/printerprofiles", methods=["GET"])
@with_revalidation_checking(etag_factory=_etag,
                            lastmodified_factory=_lastmodified,
                            unless=lambda: request.values.get("force", "false") in valid_boolean_trues)
@no_firstrun_access
@Permissions.CONNECTION.require(403)
def printerProfilesList():
	all_profiles = printerProfileManager.get_all()
	return jsonify(dict(profiles=_convert_profiles(all_profiles)))

@api.route("/printerprofiles", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def printerProfilesAdd():
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		json_data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if json_data is None:
		return make_response("Malformed JSON body in request", 400)

	if not "profile" in json_data:
		return make_response("No profile included in request", 400)

	base_profile = printerProfileManager.get_default()
	if "basedOn" in json_data and isinstance(json_data["basedOn"], basestring):
		other_profile = printerProfileManager.get(json_data["basedOn"])
		if other_profile is not None:
			base_profile = other_profile

	if "id" in base_profile:
		del base_profile["id"]
	if "name" in base_profile:
		del base_profile["name"]
	if "default" in base_profile:
		del base_profile["default"]

	new_profile = json_data["profile"]
	make_default = False
	if "default" in new_profile:
		make_default = True
		del new_profile["default"]

	profile = dict_merge(base_profile, new_profile)

	if not "id" in profile:
		return make_response("Profile does not contain mandatory 'id' field", 400)
	if not "name" in profile:
		return make_response("Profile does not contain mandatory 'name' field", 400)

	try:
		saved_profile = printerProfileManager.save(profile, allow_overwrite=False, make_default=make_default,
												   trigger_event=True)
	except InvalidProfileError:
		return make_response("Profile is invalid", 400)
	except CouldNotOverwriteError:
		return make_response("Profile {} already exists and overwriting was not allowed".format(profile["id"]), 400)
	except Exception as e:
		return make_response("Could not save profile: %s" % str(e), 500)
	else:
		return jsonify(dict(profile=_convert_profile(saved_profile)))

@api.route("/printerprofiles/<string:identifier>", methods=["GET"])
@no_firstrun_access
@Permissions.CONNECTION.require(403)
def printerProfilesGet(identifier):
	profile = printerProfileManager.get(identifier)
	if profile is None:
		return make_response("Unknown profile: %s" % identifier, 404)
	else:
		return jsonify(_convert_profile(profile))

@api.route("/printerprofiles/<string:identifier>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def printerProfilesDelete(identifier):
	current_profile = printerProfileManager.get_current()
	if current_profile and current_profile["id"] == identifier:
		return make_response("Cannot delete currently selected profile: {}".format(identifier), 409)

	default_profile = printerProfileManager.get_default()
	if default_profile and default_profile["id"] == identifier:
		return make_response("Cannot delete default profile: {}".format(identifier), 409)

	printerProfileManager.remove(identifier, trigger_event=True)
	return NO_CONTENT

@api.route("/printerprofiles/<string:identifier>", methods=["PATCH"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def printerProfilesUpdate(identifier):
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		json_data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if json_data is None:
		return make_response("Malformed JSON body in request", 400)

	if not "profile" in json_data:
		return make_response("No profile included in request", 400)

	profile = printerProfileManager.get(identifier)
	if profile is None:
		profile = printerProfileManager.get_default()

	new_profile = json_data["profile"]
	merged_profile = dict_merge(profile, new_profile)

	make_default = False
	if "default" in merged_profile:
		make_default = True
		del new_profile["default"]

		merged_profile["id"] = identifier

	try:
		saved_profile = printerProfileManager.save(merged_profile, allow_overwrite=True, make_default=make_default,
												   trigger_event=True)
	except InvalidProfileError:
		return make_response("Profile is invalid", 400)
	except CouldNotOverwriteError:
		return make_response("Profile already exists and overwriting was not allowed", 400)
	except Exception as e:
		return make_response("Could not save profile: %s" % str(e), 500)
	else:
		return jsonify(dict(profile=_convert_profile(saved_profile)))

def _convert_profiles(profiles):
	result = dict()
	for identifier, profile in profiles.items():
		result[identifier] = _convert_profile(profile)
	return result

def _convert_profile(profile):
	default = printerProfileManager.get_default()["id"]
	current = printerProfileManager.get_current_or_default()["id"]

	converted = copy.deepcopy(profile)
	converted["resource"] = url_for(".printerProfilesGet", identifier=profile["id"], _external=True)
	converted["default"] = (profile["id"] == default)
	converted["current"] = (profile["id"] == current)
	return converted



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import tornado
from tornado import escape, gen, websocket

try:
    from urllib.parse import urlparse # py3
except ImportError:
    from urlparse import urlparse # py2


class SockJSWebSocketHandler(websocket.WebSocketHandler):
    if tornado.version_info[0] == 4 and tornado.version_info[1] > 1:
        def get_compression_options(self):
            # let tornado use compression when Sec-WebSocket-Extensions:permessage-deflate is provided
            return {}

    SUPPORTED_METHODS = ('GET',)

    def check_origin(self, origin):
        # let tornado first check if connection from the same domain
        same_domain = super(SockJSWebSocketHandler, self).check_origin(origin)
        if same_domain:
            return True

        # this is cross-origin connection - check using SockJS server settings
        allow_origin = self.server.settings.get("websocket_allow_origin", "*")
        if allow_origin == "":
            return False
        elif allow_origin == "*":
            return True
        else:
            parsed_origin = urlparse(origin)
            origin = parsed_origin.netloc
            origin = origin.lower()
            return origin in allow_origin

    def abort_connection(self):
        if self.ws_connection:
            self.ws_connection._abort()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin


class VirtualPrinterPlugin(octoprint.plugin.SettingsPlugin):
	def virtual_printer_factory(self, comm_instance, port, baudrate,
	                            read_timeout):
		if not port == "VIRTUAL":
			return None

		if not self._settings.global_get_boolean(
				["devel", "virtualPrinter", "enabled"]):
			return None

		import logging.handlers
		from octoprint.logging.handlers import CleaningTimedRotatingFileHandler

		seriallog_handler = CleaningTimedRotatingFileHandler(self._settings.get_plugin_logfile_path(postfix="serial"),
		                                                     when="D",
		                                                     backupCount=3)
		seriallog_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
		seriallog_handler.setLevel(logging.DEBUG)

		from . import virtual

		serial_obj = virtual.VirtualPrinter(seriallog_handler=seriallog_handler,
		                                    read_timeout=float(read_timeout))
		return serial_obj


__plugin_name__ = "Virtual Printer"
__plugin_author__ = "Gina Häußge, based on work by Daid Braam"
__plugin_homepage__ = "https://github.com/foosel/OctoPrint/wiki/Plugin:-Virtual-Printer"
__plugin_license__ = "AGPLv3"
__plugin_description__ = "Provides a virtual printer via a virtual serial port for development and testing purposes"


def __plugin_load__():
	plugin = VirtualPrinterPlugin()

	global __plugin_implementation__
	__plugin_implementation__ = plugin

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.comm.transport.serial.factory": plugin.virtual_printer_factory
	}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, jsonify, make_response, url_for

from octoprint.filemanager.destinations import FileDestinations
from octoprint.settings import settings, valid_boolean_trues
from octoprint.server import printer, fileManager, slicingManager, eventManager, NO_CONTENT, current_user
from octoprint.server.util.flask import no_firstrun_access, get_json_command_from_request, with_revalidation_checking
from octoprint.server.api import api
from octoprint.events import Events
from octoprint.access.permissions import Permissions

import octoprint.filemanager
import octoprint.filemanager.util
import octoprint.filemanager.storage
import octoprint.slicing

from octoprint.util import sv

import os
import psutil
import hashlib
import logging
import threading


#~~ GCODE file handling

_file_cache = dict()
_file_cache_mutex = threading.RLock()

_DATA_FORMAT_VERSION = "v2"

def _clear_file_cache():
	with _file_cache_mutex:
		_file_cache.clear()

def _create_lastmodified(path, recursive):
	if path.endswith("/files"):
		# all storages involved
		lms = [0]
		for storage in fileManager.registered_storages:
			try:
				lms.append(fileManager.last_modified(storage, recursive=recursive))
			except Exception:
				logging.getLogger(__name__).exception("There was an error retrieving the last modified data from storage {}".format(storage))
				lms.append(None)

		if any(filter(lambda x: x is None, lms)):
			# we return None if ANY of the involved storages returned None
			return None

		# if we reach this point, we return the maximum of all dates
		return max(lms)

	elif path.endswith("/files/local"):
		# only local storage involved
		try:
			return fileManager.last_modified(FileDestinations.LOCAL, recursive=recursive)
		except Exception:
			logging.getLogger(__name__).exception("There was an error retrieving the last modified data from storage {}".format(FileDestinations.LOCAL))
			return None

	else:
		return None


def _create_etag(path, filter, recursive, lm=None):
	if lm is None:
		lm = _create_lastmodified(path, recursive)

	if lm is None:
		return None

	hash = hashlib.sha1()
	def hash_update(value):
		value = value.encode('utf-8')
		hash.update(value)
	hash_update(str(lm))
	hash_update(str(filter))
	hash_update(str(recursive))

	if path.endswith("/files") or path.endswith("/files/sdcard"):
		# include sd data in etag
		hash_update(repr(sorted(printer.get_sd_files(), key=lambda x: sv(x[0]))))

	hash_update(_DATA_FORMAT_VERSION) # increment version if we change the API format

	return hash.hexdigest()


@api.route("/files", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(etag_factory=lambda lm=None: _create_etag(request.path,
                                                                      request.values.get("filter", False),
                                                                      request.values.get("recursive", False),
                                                                      lm=lm),
                            lastmodified_factory=lambda: _create_lastmodified(request.path,
                                                                              request.values.get("recursive", False)),
                            unless=lambda: request.values.get("force", False) or request.values.get("_refresh", False))
def readGcodeFiles():
	filter = request.values.get("filter", False)
	recursive = request.values.get("recursive", "false") in valid_boolean_trues
	force = request.values.get("force", "false") in valid_boolean_trues

	files = _getFileList(FileDestinations.LOCAL, filter=filter, recursive=recursive, allow_from_cache=not force)
	files.extend(_getFileList(FileDestinations.SDCARD))

	usage = psutil.disk_usage(settings().getBaseFolder("uploads", check_writable=False))
	return jsonify(files=files, free=usage.free, total=usage.total)


@api.route("/files/<string:origin>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(etag_factory=lambda lm=None: _create_etag(request.path,
                                                                      request.values.get("filter", False),
                                                                      request.values.get("recursive", False),
                                                                      lm=lm),
                            lastmodified_factory=lambda: _create_lastmodified(request.path,
                                                                              request.values.get("recursive", False)),
                            unless=lambda: request.values.get("force", False) or request.values.get("_refresh", False))
def readGcodeFilesForOrigin(origin):
	if origin not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
		return make_response("Unknown origin: %s" % origin, 404)

	filter = request.values.get("filter", False)
	recursive = request.values.get("recursive", "false") in valid_boolean_trues
	force = request.values.get("force", "false") in valid_boolean_trues

	files = _getFileList(origin, filter=filter, recursive=recursive, allow_from_cache=not force)

	if origin == FileDestinations.LOCAL:
		usage = psutil.disk_usage(settings().getBaseFolder("uploads", check_writable=False))
		return jsonify(files=files, free=usage.free, total=usage.total)
	else:
		return jsonify(files=files)


def _getFileDetails(origin, path, recursive=True):
	parent, path = os.path.split(path)
	files = _getFileList(origin, path=parent, recursive=recursive)

	for f in files:
		if f["name"] == path:
			return f
	else:
		return None


def _getFileList(origin, path=None, filter=None, recursive=False, allow_from_cache=True):
	if origin == FileDestinations.SDCARD:
		sdFileList = printer.get_sd_files()

		files = []
		if sdFileList is not None:
			for sdFile, sdSize in sdFileList:
				type_path = octoprint.filemanager.get_file_type(sdFile)
				if not type_path:
					# only supported extensions
					continue
				else:
					file_type = type_path[0]

				file = {
					"type": file_type,
					"typePath": type_path,
					"name": sdFile,
					"display": sdFile,
					"path": sdFile,
					"origin": FileDestinations.SDCARD,
					"refs": {
						"resource": url_for(".readGcodeFile", target=FileDestinations.SDCARD, filename=sdFile, _external=True)
					}
				}
				if sdSize is not None:
					file.update({"size": sdSize})
				files.append(file)
	else:
		filter_func = None
		if filter:
			filter_func = lambda entry, entry_data: octoprint.filemanager.valid_file_type(entry, type=filter)

		with _file_cache_mutex:
			cache_key = "{}:{}:{}:{}".format(origin, path, recursive, filter)
			files, lastmodified = _file_cache.get(cache_key, ([], None))
			if not allow_from_cache or lastmodified is None or lastmodified < fileManager.last_modified(origin, path=path, recursive=recursive):
				files = list(fileManager.list_files(origin, path=path, filter=filter_func, recursive=recursive)[origin].values())
				lastmodified = fileManager.last_modified(origin, path=path, recursive=recursive)
				_file_cache[cache_key] = (files, lastmodified)

		def analyse_recursively(files, path=None):
			if path is None:
				path = ""

			result = []
			for file_or_folder in files:
				# make a shallow copy in order to not accidentally modify the cached data
				file_or_folder = dict(file_or_folder)

				file_or_folder["origin"] = FileDestinations.LOCAL

				if file_or_folder["type"] == "folder":
					if "children" in file_or_folder:
						file_or_folder["children"] = analyse_recursively(file_or_folder["children"].values(), path + file_or_folder["name"] + "/")

					file_or_folder["refs"] = dict(resource=url_for(".readGcodeFile", target=FileDestinations.LOCAL, filename=path + file_or_folder["name"], _external=True))
				else:
					if "analysis" in file_or_folder and octoprint.filemanager.valid_file_type(file_or_folder["name"], type="gcode"):
						file_or_folder["gcodeAnalysis"] = file_or_folder["analysis"]
						del file_or_folder["analysis"]

					if "history" in file_or_folder and octoprint.filemanager.valid_file_type(file_or_folder["name"], type="gcode"):
						# convert print log
						history = file_or_folder["history"]
						del file_or_folder["history"]
						success = 0
						failure = 0
						last = None
						for entry in history:
							success += 1 if "success" in entry and entry["success"] else 0
							failure += 1 if "success" in entry and not entry["success"] else 0
							if not last or ("timestamp" in entry and "timestamp" in last and entry["timestamp"] > last["timestamp"]):
								last = entry
						if last:
							prints = dict(
								success=success,
								failure=failure,
								last=dict(
									success=last["success"],
									date=last["timestamp"]
								)
							)
							if "printTime" in last:
								prints["last"]["printTime"] = last["printTime"]
							file_or_folder["prints"] = prints

					file_or_folder["refs"] = dict(resource=url_for(".readGcodeFile", target=FileDestinations.LOCAL, filename=file_or_folder["path"], _external=True),
					                              download=url_for("index", _external=True) + "downloads/files/" + FileDestinations.LOCAL + "/" + file_or_folder["path"])

				result.append(file_or_folder)

			return result

		files = analyse_recursively(files)

	return files


def _verifyFileExists(origin, filename):
	if origin == FileDestinations.SDCARD:
		return filename in (x[0] for x in printer.get_sd_files())
	else:
		return fileManager.file_exists(origin, filename)


def _verifyFolderExists(origin, foldername):
	if origin == FileDestinations.SDCARD:
		return False
	else:
		return fileManager.folder_exists(origin, foldername)


def _isBusy(target, path):
	currentOrigin, currentPath = _getCurrentFile()
	if currentPath is not None and currentOrigin == target and fileManager.file_in_path(FileDestinations.LOCAL, path, currentPath) and (printer.is_printing() or printer.is_paused()):
		return True

	return any(target == x[0] and fileManager.file_in_path(FileDestinations.LOCAL, path, x[1]) for x in fileManager.get_busy_files())

@api.route("/files/<string:target>", methods=["POST"])
@no_firstrun_access
@Permissions.FILES_UPLOAD.require(403)
def uploadGcodeFile(target):
	input_name = "file"
	input_upload_name = input_name + "." + settings().get(["server", "uploads", "nameSuffix"])
	input_upload_path = input_name + "." + settings().get(["server", "uploads", "pathSuffix"])
	if input_upload_name in request.values and input_upload_path in request.values:
		if not target in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
			return make_response("Unknown target: %s" % target, 404)

		upload = octoprint.filemanager.util.DiskFileWrapper(request.values[input_upload_name], request.values[input_upload_path])

		# Store any additional user data the caller may have passed.
		userdata = None
		if "userdata" in request.values:
			import json
			try:
				userdata = json.loads(request.values["userdata"])
			except Exception:
				return make_response("userdata contains invalid JSON", 400)

		if target == FileDestinations.SDCARD and not settings().getBoolean(["feature", "sdSupport"]):
			return make_response("SD card support is disabled", 404)

		sd = target == FileDestinations.SDCARD
		selectAfterUpload = "select" in request.values and request.values["select"] in valid_boolean_trues and Permissions.FILES_SELECT.can()
		printAfterSelect = "print" in request.values and request.values["print"] in valid_boolean_trues and Permissions.PRINT.can()

		if sd:
			# validate that all preconditions for SD upload are met before attempting it
			if not (printer.is_operational() and not (printer.is_printing() or printer.is_paused())):
				return make_response("Can not upload to SD card, printer is either not operational or already busy", 409)
			if not printer.is_sd_ready():
				return make_response("Can not upload to SD card, not yet initialized", 409)

		# determine future filename of file to be uploaded, abort if it can't be uploaded
		try:
			# FileDestinations.LOCAL = should normally be target, but can't because SDCard handling isn't implemented yet
			canonPath, canonFilename = fileManager.canonicalize(FileDestinations.LOCAL, upload.filename)
			futurePath = fileManager.sanitize_path(FileDestinations.LOCAL, canonPath)
			futureFilename = fileManager.sanitize_name(FileDestinations.LOCAL, canonFilename)
		except Exception:
			canonFilename = None
			futurePath = None
			futureFilename = None

		if futureFilename is None:
			return make_response("Can not upload file %s, wrong format?" % upload.filename, 415)

		if "path" in request.values and request.values["path"]:
			# we currently only support uploads to sdcard via local, so first target is local instead of "target"
			futurePath = fileManager.sanitize_path(FileDestinations.LOCAL, request.values["path"])

		# prohibit overwriting currently selected file while it's being printed
		futureFullPath = fileManager.join_path(FileDestinations.LOCAL, futurePath, futureFilename)
		futureFullPathInStorage = fileManager.path_in_storage(FileDestinations.LOCAL, futureFullPath)

		if not printer.can_modify_file(futureFullPathInStorage, sd):
			return make_response("Trying to overwrite file that is currently being printed: %s" % futureFullPath, 409)

		reselect = printer.is_current_file(futureFullPathInStorage, sd)

		user = current_user.get_name()

		def fileProcessingFinished(filename, absFilename, destination):
			"""
			Callback for when the file processing (upload, optional slicing, addition to analysis queue) has
			finished.

			Depending on the file's destination triggers either streaming to SD card or directly calls selectAndOrPrint.
			"""

			if destination == FileDestinations.SDCARD and octoprint.filemanager.valid_file_type(filename, "machinecode"):
				return filename, printer.add_sd_file(filename, absFilename,
				                                     on_success=selectAndOrPrint,
				                                     tags={"source:api", "api:files.sd"})
			else:
				selectAndOrPrint(filename, absFilename, destination)
				return filename

		def selectAndOrPrint(filename, absFilename, destination):
			"""
			Callback for when the file is ready to be selected and optionally printed. For SD file uploads this is only
			the case after they have finished streaming to the printer, which is why this callback is also used
			for the corresponding call to addSdFile.

			Selects the just uploaded file if either selectAfterUpload or printAfterSelect are True, or if the
			exact file is already selected, such reloading it.
			"""
			if octoprint.filemanager.valid_file_type(added_file, "gcode") and (selectAfterUpload or printAfterSelect or reselect):
				printer.select_file(absFilename, destination == FileDestinations.SDCARD, printAfterSelect, user)

		try:
			added_file = fileManager.add_file(FileDestinations.LOCAL, futureFullPathInStorage, upload,
			                                  allow_overwrite=True,
			                                  display=canonFilename)
		except octoprint.filemanager.storage.StorageError as e:
			if e.code == octoprint.filemanager.storage.StorageError.INVALID_FILE:
				return make_response("Could not upload the file \"{}\", invalid type".format(upload.filename), 400)
			else:
				return make_response("Could not upload the file \"{}\"".format(upload.filename), 500)

		if octoprint.filemanager.valid_file_type(added_file, "stl"):
			filename = added_file
			done = True
		else:
			filename = fileProcessingFinished(added_file, fileManager.path_on_disk(FileDestinations.LOCAL, added_file), target)
			done = not sd

		if userdata is not None:
			# upload included userdata, add this now to the metadata
			fileManager.set_additional_metadata(FileDestinations.LOCAL, added_file, "userdata", userdata)

		sdFilename = None
		if isinstance(filename, tuple):
			filename, sdFilename = filename

		eventManager.fire(Events.UPLOAD, {"name": futureFilename,
		                                  "path": filename,
		                                  "target": target})

		files = {}
		location = url_for(".readGcodeFile", target=FileDestinations.LOCAL, filename=filename, _external=True)
		files.update({
			FileDestinations.LOCAL: {
				"name": futureFilename,
				"path": filename,
				"origin": FileDestinations.LOCAL,
				"refs": {
					"resource": location,
					"download": url_for("index", _external=True) + "downloads/files/" + FileDestinations.LOCAL + "/" + filename
				}
			}
		})

		if sd and sdFilename:
			location = url_for(".readGcodeFile", target=FileDestinations.SDCARD, filename=sdFilename, _external=True)
			files.update({
				FileDestinations.SDCARD: {
					"name": sdFilename,
					"path": sdFilename,
					"origin": FileDestinations.SDCARD,
					"refs": {
						"resource": location
					}
				}
			})

		r = make_response(jsonify(files=files, done=done), 201)
		r.headers["Location"] = location
		return r

	elif "foldername" in request.values:
		foldername = request.values["foldername"]

		if not target in [FileDestinations.LOCAL]:
			return make_response("Unknown target: %s" % target, 400)

		canonPath, canonName = fileManager.canonicalize(target, foldername)
		futurePath = fileManager.sanitize_path(target, canonPath)
		futureName = fileManager.sanitize_name(target, canonName)
		if not futureName or not futurePath:
			return make_response("Can't create a folder with an empty name", 400)

		if "path" in request.values and request.values["path"]:
			futurePath = fileManager.sanitize_path(FileDestinations.LOCAL,
			                                       request.values["path"])

		futureFullPath = fileManager.join_path(target, futurePath, futureName)
		if octoprint.filemanager.valid_file_type(futureName):
			return make_response("Can't create a folder named %s, please try another name" % futureName, 409)

		try:
			added_folder = fileManager.add_folder(target, futureFullPath, display=canonName)
		except octoprint.filemanager.storage.StorageError as e:
			if e.code == octoprint.filemanager.storage.StorageError.INVALID_DIRECTORY:
				return make_response("Could not create folder {}, invalid directory".format(futureName))
			else:
				return make_response("Could not create folder {}".format(futureName))

		location = url_for(".readGcodeFile",
		                   target=FileDestinations.LOCAL,
		                   filename=added_folder,
		                   _external=True)
		folder = dict(name=futureName,
		              path=added_folder,
		              origin=target,
		              refs=dict(resource=location))

		r = make_response(jsonify(folder=folder, done=True), 201)
		r.headers["Location"] = location
		return r

	else:
		return make_response("No file to upload and no folder to create", 400)


@api.route("/files/<string:target>/<path:filename>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
def readGcodeFile(target, filename):
	if not target in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
		return make_response("Unknown target: %s" % target, 404)

	recursive = False
	if "recursive" in request.values:
		recursive = request.values["recursive"] in valid_boolean_trues

	file = _getFileDetails(target, filename, recursive=recursive)
	if not file:
		return make_response("File not found on '%s': %s" % (target, filename), 404)

	return jsonify(file)


@api.route("/files/<string:target>/<path:filename>", methods=["POST"])
@no_firstrun_access
def gcodeFileCommand(filename, target):
	if not target in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
		return make_response("Unknown target: %s" % target, 404)

	# valid file commands, dict mapping command name to mandatory parameters
	valid_commands = {
		"select": [],
		"slice": [],
		"analyse": [],
		"copy": ["destination"],
		"move": ["destination"]
	}

	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	user = current_user.get_name()

	if command == "select":
		with Permissions.FILES_SELECT.require(403):
			if not _verifyFileExists(target, filename):
				return make_response("File not found on '%s': %s" % (target, filename), 404)

			# selects/loads a file
			if not octoprint.filemanager.valid_file_type(filename, type="machinecode"):
				return make_response("Cannot select {filename} for printing, not a machinecode file".format(**locals()), 415)

			printAfterLoading = False
			if "print" in data and data["print"] in valid_boolean_trues:
				with Permissions.PRINT.require(403):
					if not printer.is_operational():
						return make_response("Printer is not operational, cannot directly start printing", 409)
					printAfterLoading = True

			sd = False
			if target == FileDestinations.SDCARD:
				filenameToSelect = filename
				sd = True
			else:
				filenameToSelect = fileManager.path_on_disk(target, filename)
			printer.select_file(filenameToSelect, sd, printAfterLoading, user)

	elif command == "slice":
		with Permissions.SLICE.require(403):
			if not _verifyFileExists(target, filename):
				return make_response("File not found on '%s': %s" % (target, filename), 404)

			try:
				if "slicer" in data:
					slicer = data["slicer"]
					del data["slicer"]
					slicer_instance = slicingManager.get_slicer(slicer)

				elif "cura" in slicingManager.registered_slicers:
					slicer = "cura"
					slicer_instance = slicingManager.get_slicer("cura")

				else:
					return make_response("Cannot slice {filename}, no slicer available".format(**locals()), 415)
			except octoprint.slicing.UnknownSlicer as e:
				return make_response("Slicer {slicer} is not available".format(slicer=e.slicer), 400)

			if not any([octoprint.filemanager.valid_file_type(filename, type=source_file_type) for source_file_type in slicer_instance.get_slicer_properties().get("source_file_types", ["model"])]):
				return make_response("Cannot slice {filename}, not a model file".format(**locals()), 415)

			cores = psutil.cpu_count()
			if slicer_instance.get_slicer_properties().get("same_device", True) and (printer.is_printing() or printer.is_paused()) and (cores is None or cores < 2):
				# slicer runs on same device as OctoPrint, slicing while printing is hence disabled
				return make_response("Cannot slice on {slicer} while printing on single core systems or systems of unknown core count due to performance reasons".format(**locals()), 409)

			if "destination" in data and data["destination"]:
				destination = data["destination"]
				del data["destination"]
			elif "gcode" in data and data["gcode"]:
				destination = data["gcode"]
				del data["gcode"]
			else:
				import os
				name, _ = os.path.splitext(filename)
				destination = name + "." + slicer_instance.get_slicer_properties().get("destination_extensions", ["gco", "gcode", "g"])[0]

			full_path = destination
			if "path" in data and data["path"]:
				full_path = fileManager.join_path(target, data["path"], destination)
			else:
				path, _ = fileManager.split_path(target, filename)
				if path:
					full_path = fileManager.join_path(target, path, destination)

			canon_path, canon_name = fileManager.canonicalize(target, full_path)
			sanitized_name = fileManager.sanitize_name(target, canon_name)

			if canon_path:
				full_path = fileManager.join_path(target, canon_path, sanitized_name)
			else:
				full_path = sanitized_name

			# prohibit overwriting the file that is currently being printed
			currentOrigin, currentFilename = _getCurrentFile()
			if currentFilename == full_path and currentOrigin == target and (printer.is_printing() or printer.is_paused()):
				make_response("Trying to slice into file that is currently being printed: %s" % full_path, 409)

			if "profile" in data and data["profile"]:
				profile = data["profile"]
				del data["profile"]
			else:
				profile = None

			if "printerProfile" in data and data["printerProfile"]:
				printerProfile = data["printerProfile"]
				del data["printerProfile"]
			else:
				printerProfile = None

			if "position" in data and data["position"] and isinstance(data["position"], dict) and "x" in data["position"] and "y" in data["position"]:
				position = data["position"]
				del data["position"]
			else:
				position = None

			select_after_slicing = False
			if "select" in data and data["select"] in valid_boolean_trues:
				if not printer.is_operational():
					return make_response("Printer is not operational, cannot directly select for printing", 409)
				select_after_slicing = True

			print_after_slicing = False
			if "print" in data and data["print"] in valid_boolean_trues:
				if not printer.is_operational():
					return make_response("Printer is not operational, cannot directly start printing", 409)
				select_after_slicing = print_after_slicing = True

			override_keys = [k for k in data if k.startswith("profile.") and data[k] is not None]
			overrides = dict()
			for key in override_keys:
				overrides[key[len("profile."):]] = data[key]

			def slicing_done(target, path, select_after_slicing, print_after_slicing):
				if select_after_slicing or print_after_slicing:
					sd = False
					if target == FileDestinations.SDCARD:
						filenameToSelect = path
						sd = True
					else:
						filenameToSelect = fileManager.path_on_disk(target, path)
					printer.select_file(filenameToSelect, sd, print_after_slicing, user)

			try:
				fileManager.slice(slicer, target, filename, target, full_path,
				                  profile=profile,
				                  printer_profile_id=printerProfile,
				                  position=position,
				                  overrides=overrides,
								  display=canon_name,
				                  callback=slicing_done,
				                  callback_args=(target, full_path, select_after_slicing, print_after_slicing))
			except octoprint.slicing.UnknownProfile:
				return make_response("Profile {profile} doesn't exist".format(**locals()), 400)

			files = {}
			location = url_for(".readGcodeFile", target=target, filename=full_path, _external=True)
			result = {
				"name": destination,
				"path": full_path,
				"display": canon_name,
				"origin": FileDestinations.LOCAL,
				"refs": {
					"resource": location,
					"download": url_for("index", _external=True) + "downloads/files/" + target + "/" + full_path
				}
			}

			r = make_response(jsonify(result), 202)
			r.headers["Location"] = location
			return r

	elif command == "analyse":
		with Permissions.FILES_UPLOAD.require(403):
			if not _verifyFileExists(target, filename):
				return make_response("File not found on '%s': %s" % (target, filename), 404)

			printer_profile = None
			if "printerProfile" in data and data["printerProfile"]:
				printer_profile = data["printerProfile"]

			if not fileManager.analyse(target, filename, printer_profile_id=printer_profile):
				return make_response("No analysis possible for {} on {}".format(filename, target), 400)

	elif command == "copy" or command == "move":
		with Permissions.FILES_UPLOAD.require(403):
			# Copy and move are only possible on local storage
			if not target in [FileDestinations.LOCAL]:
				return make_response("Unsupported target for {}: {}".format(command, target), 400)

			if not _verifyFileExists(target, filename) and not _verifyFolderExists(target, filename):
				return make_response("File or folder not found on {}: {}".format(target, filename), 404)

			path, name = fileManager.split_path(target, filename)

			destination = data["destination"]
			dst_path, dst_name = fileManager.split_path(target, destination)
			sanitized_destination = fileManager.join_path(target, dst_path, fileManager.sanitize_name(target, dst_name))

			if _verifyFolderExists(target, destination) and sanitized_destination != filename:
				# destination is an existing folder and not ourselves (= display rename), we'll assume we are supposed
				# to move filename to this folder under the same name
				destination = fileManager.join_path(target, destination, name)

			if _verifyFileExists(target, destination) or _verifyFolderExists(target, destination):
				return make_response("File or folder does already exist on {}: {}".format(target, destination), 409)

			is_file = fileManager.file_exists(target, filename)
			is_folder = fileManager.folder_exists(target, filename)

			if not (is_file or is_folder):
				return make_response("{} on {} is neither file or folder, can't {}".format(filename, target, command), 400)

			if command == "copy":
				# destination already there? error...
				if _verifyFileExists(target, destination) or _verifyFolderExists(target, destination):
					return make_response("File or folder does already exist on {}: {}".format(target, destination), 409)

				if is_file:
					fileManager.copy_file(target, filename, destination)
				else:
					fileManager.copy_folder(target, filename, destination)

			elif command == "move":
				if _isBusy(target, filename):
					return make_response("Trying to move a file or folder that is currently in use: {}".format(filename), 409)

				# destination already there AND not ourselves (= display rename)? error...
				if (_verifyFileExists(target, destination) or _verifyFolderExists(target, destination)) \
						and sanitized_destination != filename:
					return make_response("File or folder does already exist on {}: {}".format(target, destination), 409)

				# deselect the file if it's currently selected
				currentOrigin, currentFilename = _getCurrentFile()
				if currentFilename is not None and filename == currentFilename:
					printer.unselect_file()

				if is_file:
					fileManager.move_file(target, filename, destination)
				else:
					fileManager.move_folder(target, filename, destination)

			location = url_for(".readGcodeFile", target=target, filename=destination, _external=True)
			result = {
				"name": name,
				"path": destination,
				"origin": FileDestinations.LOCAL,
				"refs": {
					"resource": location
				}
			}
			if is_file:
				result["refs"]["download"] = url_for("index", _external=True) + "downloads/files/" + target + "/" + destination

			r = make_response(jsonify(result), 201)
			r.headers["Location"] = location
			return r

	return NO_CONTENT


@api.route("/files/<string:target>/<path:filename>", methods=["DELETE"])
@no_firstrun_access
@Permissions.FILES_DELETE.require(403)
def deleteGcodeFile(filename, target):
	if not _verifyFileExists(target, filename) and not _verifyFolderExists(target, filename):
		return make_response("File/Folder not found on '%s': %s" % (target, filename), 404)

	if _verifyFileExists(target, filename):
		if not target in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
			return make_response("Unknown target: %s" % target, 400)

		if _isBusy(target, filename):
			return make_response("Trying to delete a file that is currently in use: %s" % filename, 409)

		# deselect the file if it's currently selected
		currentOrigin, currentPath = _getCurrentFile()
		if currentPath is not None and currentOrigin == target and filename == currentPath:
			printer.unselect_file()

		# delete it
		if target == FileDestinations.SDCARD:
			printer.delete_sd_file(filename, tags={"source:api", "api:files.sd"})
		else:
			fileManager.remove_file(target, filename)

	elif _verifyFolderExists(target, filename):
		if not target in [FileDestinations.LOCAL]:
			return make_response("Unknown target: %s" % target, 400)

		if _isBusy(target, filename):
			return make_response("Trying to delete a folder that contains a file that is currently in use: %s" % filename, 409)

		# deselect the file if it's currently selected
		currentOrigin, currentPath = _getCurrentFile()
		if currentPath is not None and currentOrigin == target and fileManager.file_in_path(target, filename, currentPath):
			printer.unselect_file()

		# delete it
		fileManager.remove_folder(target, filename, recursive=True)

	return NO_CONTENT

def _getCurrentFile():
	currentJob = printer.get_current_job()
	if currentJob is not None and "file" in currentJob and "path" in currentJob["file"] and "origin" in currentJob["file"]:
		return currentJob["file"]["origin"], currentJob["file"]["path"]
	else:
		return None, None


class WerkzeugFileWrapper(octoprint.filemanager.util.AbstractFileWrapper):
	"""
	A wrapper around a Werkzeug ``FileStorage`` object.

	Arguments:
	    file_obj (werkzeug.datastructures.FileStorage): The Werkzeug ``FileStorage`` instance to wrap.

	.. seealso::

	   `werkzeug.datastructures.FileStorage <http://werkzeug.pocoo.org/docs/0.10/datastructures/#werkzeug.datastructures.FileStorage>`_
	        The documentation of Werkzeug's ``FileStorage`` class.
	"""
	def __init__(self, file_obj):
		octoprint.filemanager.util.AbstractFileWrapper.__init__(self, file_obj.filename)
		self.file_obj = file_obj

	def save(self, path):
		"""
		Delegates to ``werkzeug.datastructures.FileStorage.save``
		"""
		self.file_obj.save(path)

	def stream(self):
		"""
		Returns ``werkzeug.datastructures.FileStorage.stream``
		"""
		return self.file_obj.stream



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from octoprint.events import Events
from octoprint.access import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.util import to_unicode

from .checks.firmware_unsafe import FirmwareUnsafeChecks

import flask
from flask_babel import gettext

import textwrap

TERMINAL_SAFETY_WARNING = """
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
{message}

Learn more at https://faq.octoprint.org/warning-{warning_type}
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

"""

SAFETY_CHECKS = {
	"firmware-unsafe": FirmwareUnsafeChecks.as_dict()
}

class PrinterSafetyCheckPlugin(octoprint.plugin.AssetPlugin,
                               octoprint.plugin.EventHandlerPlugin,
                               octoprint.plugin.SimpleApiPlugin,
                               octoprint.plugin.TemplatePlugin):

	# noinspection PyMissingConstructor
	def __init__(self):
		self._warnings = dict()
		self._scan_received = True

	##~~ TemplatePlugin API

	def get_template_configs(self):
		return [
			dict(type="sidebar",
			     name=gettext("Printer Safety Warning"),
			     data_bind="visible: printerState.isOperational() && loginState.isAdmin() && warnings().length > 0",
			     icon="exclamation-triangle",
			     styles_wrapper=["display: none"])
		]

	##~~ AssetPlugin API

	def get_assets(self):
		return dict(js=("js/printer_safety_check.js",),
		            clientjs=("clientjs/printer_safety_check.js",),
		            css=("css/printer_safety_check.css",),
		            less=("less/printer_safety_check.less",))

	##~~ EventHandlerPlugin API

	def on_event(self, event, payload):
		if event == Events.DISCONNECTED:
			self._reset_warnings()
			self._reset_state()
			self._reset_checks()

	##~~ SimpleApiPlugin API

	def on_api_get(self, request):
		if not Permissions.PLUGIN_PRINTER_SAFETY_CHECK_DISPLAY.can():
			return flask.make_response("Insufficient rights", 403)
		return flask.jsonify(self._warnings)

	##~~ GCODE received hook handler

	def on_gcode_received(self, comm_instance, line, *args, **kwargs):
		if self._scan_received:
			self._run_checks("received", to_unicode(line, errors="replace"))
		return line

	##~~ Firmware info hook handler

	def on_firmware_info_received(self, comm_instance, firmware_name, firmware_data):
		self._run_checks("m115",
		                 to_unicode(firmware_name, errors="replace"),
		                 dict((to_unicode(key, errors="replace"), to_unicode(value, errors="replace"))
		                      for key, value in firmware_data.items()))
		self._scan_received = False

	##~~ Firmware capability hook handler

	def on_firmware_cap_received(self, comm_instance, cap, enabled, all_caps):
		self._run_checks("cap",
		                 to_unicode(cap, errors="replace"),
		                 enabled)

	##~~ Additional permissions hook handler

	def get_additional_permissions(self):
		return [
			dict(key="DISPLAY",
			     name="Display printer safety warnings",
			     description=gettext("Allows to see printer safety warnings"),
			     roles=["display"],
			     default_groups=[USER_GROUP])
		]

	##~~ Helpers

	def _run_checks(self, check_type, *args, **kwargs):
		changes = False

		for warning_type, check_data in SAFETY_CHECKS.items():
			checks = check_data.get("checks")
			message = check_data.get("message")
			if not checks or not message:
				continue

			for check in checks:
				if not check.active:
					# skip non active checks
					continue

				method = getattr(check, check_type, None)
				if not callable(method):
					# skip uncallable checks
					continue

				# execute method
				try:
					method(*args, **kwargs)
				except Exception:
					self._logger.exception("There was an error running method {} on check {!r}".format(check_type, check))
					continue

				# check if now triggered
				if check.triggered:
					self._register_warning(warning_type, message)

					# noinspection PyUnresolvedReferences
					self._event_bus.fire(Events.PLUGIN_PRINTER_SAFETY_CHECK_WARNING, dict(check_name=check.name,
					                                                                      warning_type=warning_type))
					changes = True
					break

		if changes:
			self._ping_clients()

	def _register_warning(self, warning_type, message):
		self._log_to_terminal(TERMINAL_SAFETY_WARNING.format(message="\n".join(textwrap.wrap(message, 75)),
		                                                     warning_type=warning_type))
		self._warnings[warning_type] = message

	def _reset_warnings(self):
		self._warnings.clear()
		self._ping_clients()

	def _reset_state(self):
		self._scan_received = True

	def _reset_checks(self):
		for warning_type, check_data in SAFETY_CHECKS.items():
			checks = check_data.get("checks")
			if not checks:
				continue

			for check in checks:
				check.reset()

	def _log_to_terminal(self, message):
		if self._printer:
			lines = message.split("\n")
			self._printer.log_lines(*lines)

	def _ping_clients(self):
		self._plugin_manager.send_plugin_message(self._identifier, dict(type="update"))


def register_custom_events(*args, **kwargs):
	return ["warning",]


__plugin_name__ = "Printer Safety Check"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "http://docs.octoprint.org/en/master/bundledplugins/printer_safety_check.html"
__plugin_description__ = "Checks for unsafe printers/printer firmwares"
__plugin_disabling_discouraged__ = gettext("Without this plugin OctoPrint will no longer be able to "
                                           "check if the printer it is connected to has a known safety "
                                           "issue and inform you about that fact.")
__plugin_license__ = "AGPLv3"
__plugin_implementation__ = PrinterSafetyCheckPlugin()
__plugin_hooks__ = {
	"octoprint.comm.protocol.gcode.received": __plugin_implementation__.on_gcode_received,
	"octoprint.comm.protocol.firmware.info": __plugin_implementation__.on_firmware_info_received,
	"octoprint.comm.protocol.firmware.capabilities": __plugin_implementation__.on_firmware_cap_received,
	"octoprint.events.register_custom_events": register_custom_events,
	"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
}




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

from past.builtins import basestring

import octoprint.plugin
import octoprint.plugin.core

from octoprint.settings import valid_boolean_trues
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking, check_etag
from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions
from octoprint.util import to_bytes
from octoprint.util.pip import LocalPipCaller
from octoprint.util.version import get_octoprint_version_string, get_octoprint_version, is_octoprint_compatible, is_python_compatible
from octoprint.util.platform import get_os, is_os_compatible
from octoprint.events import Events

from flask import jsonify, make_response
from flask_babel import gettext

import io
import logging
import sarge
import sys
import requests
import re
import os
import copy
import time
import threading

from datetime import datetime

try:
	from os import scandir
except:
	from scandir import scandir

_DATA_FORMAT_VERSION = "v3"


def map_repository_entry(entry):
	if not isinstance(entry, dict):
		return None

	result = copy.deepcopy(entry)

	if not "follow_dependency_links" in result:
		result["follow_dependency_links"] = False

	result["is_compatible"] = dict(
		octoprint=True,
		os=True,
		python=True
	)

	if "compatibility" in entry:
		if "octoprint" in entry["compatibility"] and entry["compatibility"]["octoprint"] is not None and isinstance(
			entry["compatibility"]["octoprint"], (list, tuple)) and len(entry["compatibility"]["octoprint"]):
			result["is_compatible"]["octoprint"] = is_octoprint_compatible(*entry["compatibility"]["octoprint"])

		if "os" in entry["compatibility"] and entry["compatibility"]["os"] is not None and isinstance(
			entry["compatibility"]["os"], (list, tuple)) and len(entry["compatibility"]["os"]):
			result["is_compatible"]["os"] = is_os_compatible(entry["compatibility"]["os"])

		if "python" in entry["compatibility"] and entry["compatibility"]["python"] is not None and isinstance(entry["compatibility"]["python"], basestring):
			result["is_compatible"]["python"] = is_python_compatible(entry["compatibility"]["python"])
		else:
			# we default to only assume py2 compatiblity for now
			result["is_compatible"]["python"] = is_python_compatible(">=2.7,<3")

	return result


already_installed_string = "Requirement already satisfied (use --upgrade to upgrade)"
success_string = "Successfully installed"
failure_string = "Could not install"


class PluginManagerPlugin(octoprint.plugin.SimpleApiPlugin,
                          octoprint.plugin.TemplatePlugin,
                          octoprint.plugin.AssetPlugin,
                          octoprint.plugin.SettingsPlugin,
                          octoprint.plugin.StartupPlugin,
                          octoprint.plugin.BlueprintPlugin,
                          octoprint.plugin.EventHandlerPlugin):

	ARCHIVE_EXTENSIONS = (".zip", ".tar.gz", ".tgz", ".tar")

	# valid pip install URL schemes according to https://pip.pypa.io/en/stable/reference/pip_install/
	URL_SCHEMES = ("http", "https", "git",
	               "git+http", "git+https", "git+ssh", "git+git",
	               "hg+http", "hg+https", "hg+static-http", "hg+ssh",
	               "svn", "svn+svn", "svn+http", "svn+https", "svn+ssh",
	               "bzr+http", "bzr+https", "bzr+ssh", "bzr+sftp", "bzr+ftp", "bzr+lp")

	OPERATING_SYSTEMS = dict(windows=["win32"],
	                         linux=lambda x: x.startswith("linux"),
	                         macos=["darwin"],
	                         freebsd=lambda x: x.startswith("freebsd"))

	PIP_INAPPLICABLE_ARGUMENTS = dict(uninstall=["--user"])

	RECONNECT_HOOKS = ["octoprint.comm.protocol.*",]

	# noinspection PyMissingConstructor
	def __init__(self):
		self._pending_enable = set()
		self._pending_disable = set()
		self._pending_install = set()
		self._pending_uninstall = set()

		self._pip_caller = None

		self._repository_available = False
		self._repository_plugins = []
		self._repository_cache_path = None
		self._repository_cache_ttl = 0
		self._repository_mtime = None

		self._notices = dict()
		self._notices_available = False
		self._notices_cache_path = None
		self._notices_cache_ttl = 0
		self._notices_mtime = None

		self._orphans = None

		self._console_logger = None

		self._get_throttled = lambda: False

	def initialize(self):
		self._console_logger = logging.getLogger("octoprint.plugins.pluginmanager.console")
		self._repository_cache_path = os.path.join(self.get_plugin_data_folder(), "plugins.json")
		self._repository_cache_ttl = self._settings.get_int(["repository_ttl"]) * 60
		self._notices_cache_path = os.path.join(self.get_plugin_data_folder(), "notices.json")
		self._notices_cache_ttl = self._settings.get_int(["notices_ttl"]) * 60
		self._confirm_disable = self._settings.global_get_boolean(["confirm_disable"])

		self._pip_caller = LocalPipCaller(force_user=self._settings.get_boolean(["pip_force_user"]))
		self._pip_caller.on_log_call = self._log_call
		self._pip_caller.on_log_stdout = self._log_stdout
		self._pip_caller.on_log_stderr = self._log_stderr

	##~~ Body size hook

	def increase_upload_bodysize(self, current_max_body_sizes, *args, **kwargs):
		# set a maximum body size of 50 MB for plugin archive uploads
		return [("POST", r"/upload_archive", 50 * 1024 * 1024)]

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="MANAGE",
			     name="Manage plugins",
			     description=gettext("Allows to enable, disable and uninstall installed plugins."),
			     default_groups=[ADMIN_GROUP],
			     roles=["manage"]),
			dict(key="INSTALL",
			     name="Install new plugins",
			     description=gettext("Allows to install new plugins. Includes the \"Manage plugins\" permission."),
			     default_groups=[ADMIN_GROUP],
			     roles=["install"],
			     permissions=["PLUGIN_PLUGINMANAGER_MANAGE"],
			     dangerous=True)
		]

	##~~ StartupPlugin

	def on_after_startup(self):
		from octoprint.logging.handlers import CleaningTimedRotatingFileHandler
		console_logging_handler = CleaningTimedRotatingFileHandler(self._settings.get_plugin_logfile_path(postfix="console"), when="D", backupCount=3)
		console_logging_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
		console_logging_handler.setLevel(logging.DEBUG)

		self._console_logger.addHandler(console_logging_handler)
		self._console_logger.setLevel(logging.DEBUG)
		self._console_logger.propagate = False

		helpers = self._plugin_manager.get_helpers("pi_support", "get_throttled")
		if helpers and "get_throttled" in helpers:
			self._get_throttled = helpers["get_throttled"]
			if self._settings.get_boolean(["ignore_throttled"]):
				self._logger.warn("!!! THROTTLE STATE IGNORED !!! You have configured the Plugin Manager plugin to ignore an active throttle state of the underlying system. You might run into stability issues or outright corrupt your install. Consider fixing the throttling issue instead of suppressing it.")

		# decouple repository fetching from server startup
		self._fetch_all_data(do_async=True)

	##~~ SettingsPlugin

	def get_settings_defaults(self):
		return dict(
			repository="https://plugins.octoprint.org/plugins.json",
			repository_ttl=24*60,
			notices="https://plugins.octoprint.org/notices.json",
			notices_ttl=6*60,
			pip_args=None,
			pip_force_user=False,
			confirm_disable=False,
			dependency_links=False,
			hidden=[],
			ignore_throttled=False
		)

	def on_settings_save(self, data):
		octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

		self._repository_cache_ttl = self._settings.get_int(["repository_ttl"]) * 60
		self._notices_cache_ttl = self._settings.get_int(["notices_ttl"]) * 60
		self._pip_caller.force_user = self._settings.get_boolean(["pip_force_user"])
		self._confirm_disable = self._settings.global_get_boolean(["confirm_disable"])

	##~~ AssetPlugin

	def get_assets(self):
		return dict(
			js=["js/pluginmanager.js"],
			clientjs=["clientjs/pluginmanager.js"],
			css=["css/pluginmanager.css"],
			less=["less/pluginmanager.less"]
		)

	##~~ TemplatePlugin

	def get_template_configs(self):
		return [
			dict(type="settings", name=gettext("Plugin Manager"), template="pluginmanager_settings.jinja2", custom_bindings=True),
			dict(type="about", name="Plugin Licenses", template="pluginmanager_about.jinja2")
		]

	def get_template_vars(self):
		plugins = sorted(self._get_plugins(), key=lambda x: x["name"].lower())
		return dict(
			all=plugins,
			thirdparty=list(filter(lambda p: not p["bundled"], plugins)),
			archive_extensions=self.__class__.ARCHIVE_EXTENSIONS
		)

	def get_template_types(self, template_sorting, template_rules, *args, **kwargs):
		return [
			("about_thirdparty", dict(), dict(template=lambda x: x + "_about_thirdparty.jinja2"))
		]

	##~~ BlueprintPlugin

	@octoprint.plugin.BlueprintPlugin.route("/upload_archive", methods=["POST"])
	@no_firstrun_access
	@Permissions.PLUGIN_PLUGINMANAGER_INSTALL.require(403)
	def upload_archive(self):
		import flask

		input_name = "file"
		input_upload_path = input_name + "." + self._settings.global_get(["server", "uploads", "pathSuffix"])
		input_upload_name = input_name + "." + self._settings.global_get(["server", "uploads", "nameSuffix"])

		if input_upload_path not in flask.request.values or input_upload_name not in flask.request.values:
			return flask.make_response("No file included", 400)
		upload_path = flask.request.values[input_upload_path]
		upload_name = flask.request.values[input_upload_name]

		exts = list(filter(lambda x: upload_name.lower().endswith(x), self.__class__.ARCHIVE_EXTENSIONS))
		if not len(exts):
			return flask.make_response("File doesn't have a valid extension for a plugin archive", 400)

		ext = exts[0]

		import tempfile
		import shutil
		import os

		archive = tempfile.NamedTemporaryFile(delete=False, suffix="{ext}".format(**locals()))
		try:
			archive.close()
			shutil.copy(upload_path, archive.name)
			return self.command_install(path=archive.name, force="force" in flask.request.values and flask.request.values["force"] in valid_boolean_trues)
		finally:
			try:
				os.remove(archive.name)
			except Exception as e:
				self._logger.warning("Could not remove temporary file {path} again: {message}".format(path=archive.name, message=str(e)))

	def is_blueprint_protected(self):
		return False

	##~~ EventHandlerPlugin

	def on_event(self, event, payload):
		from octoprint.events import Events
		if event != Events.CONNECTIVITY_CHANGED or not payload or not payload.get("new", False):
			return
		self._fetch_all_data(do_async=True)

	##~~ SimpleApiPlugin

	def get_api_commands(self):
		return {
			"install": ["url"],
			"uninstall": ["plugin"],
			"enable": ["plugin"],
			"disable": ["plugin"],
			"cleanup": ["plugin"],
			"cleanup_all": [],
			"refresh_repository": []
		}

	def on_api_get(self, request):
		if not Permissions.PLUGIN_PLUGINMANAGER_MANAGE.can():
			return make_response("Insufficient rights", 403)

		from octoprint.server import safe_mode

		refresh_repository = request.values.get("refresh_repository", "false") in valid_boolean_trues
		if refresh_repository or not self._is_repository_cache_valid():
			self._repository_available = self._refresh_repository()

		refresh_notices = request.values.get("refresh_notices", "false") in valid_boolean_trues
		if refresh_notices or not self._is_notices_cache_valid():
			self._notices_available = self._refresh_notices()

		refresh_orphan = request.values.get("refresh_orphans", "false") in valid_boolean_trues

		def view():
			return jsonify(plugins=self._get_plugins(),
			               repository=dict(
			                   available=self._repository_available,
			                   plugins=self._repository_plugins
			               ),
			               orphan_data=self._get_orphans(refresh=refresh_orphan),
			               os=get_os(),
			               octoprint=get_octoprint_version_string(),
			               pip=dict(
			                   available=self._pip_caller.available,
			                   version=self._pip_caller.version_string,
			                   install_dir=self._pip_caller.install_dir,
			                   use_user=self._pip_caller.use_user,
			                   virtual_env=self._pip_caller.virtual_env,
			                   additional_args=self._settings.get(["pip_args"]),
			                   python=sys.executable
		                    ),
			               safe_mode=safe_mode,
			               online=self._connectivity_checker.online)

		def etag():
			import hashlib
			hash = hashlib.sha1()
			def hash_update(value):
				value = value.encode('utf-8')
				hash.update(value)

			hash_update(repr(self._get_plugins()))
			hash_update(str(self._repository_available))
			hash_update(repr(self._repository_plugins))
			hash_update(str(self._notices_available))
			hash_update(repr(self._notices))
			hash_update(repr(self._get_orphans()))
			hash_update(repr(safe_mode))
			hash_update(repr(self._connectivity_checker.online))
			hash_update(repr(_DATA_FORMAT_VERSION))
			return hash.hexdigest()

		def condition():
			return check_etag(etag())

		return with_revalidation_checking(etag_factory=lambda *args, **kwargs: etag(),
		                                  condition=lambda *args, **kwargs: condition(),
		                                  unless=lambda: refresh_repository or refresh_notices or refresh_orphan)(view)()

	def on_api_command(self, command, data):
		if not Permissions.PLUGIN_PLUGINMANAGER_MANAGE.can():
			return make_response("Insufficient rights", 403)

		if self._printer.is_printing() or self._printer.is_paused():
			# do not update while a print job is running
			return make_response("Printer is currently printing or paused", 409)

		if command == "install":
			if not Permissions.PLUGIN_PLUGINMANAGER_INSTALL.can():
				return make_response("Insufficient rights", 403)
			url = data["url"]
			plugin_name = data["plugin"] if "plugin" in data else None
			return self.command_install(url=url,
			                            force="force" in data and data["force"] in valid_boolean_trues,
			                            dependency_links="dependency_links" in data
			                                             and data["dependency_links"] in valid_boolean_trues,
			                            reinstall=plugin_name)

		elif command == "uninstall":
			plugin_name = data["plugin"]
			if not plugin_name in self._plugin_manager.plugins:
				return make_response("Unknown plugin: %s" % plugin_name, 404)

			plugin = self._plugin_manager.plugins[plugin_name]
			return self.command_uninstall(plugin, cleanup=data.get("cleanup", False))

		elif command == "cleanup":
			plugin = data["plugin"]
			try:
				plugin = self._plugin_manager.plugins[plugin]
			except KeyError:
				# not installed, we are cleaning up left overs, that's ok
				pass

			return self.command_cleanup(plugin, include_disabled=True)

		elif command == "cleanup_all":
			return self.command_cleanup_all()

		elif command == "enable" or command == "disable":
			plugin_name = data["plugin"]
			if not plugin_name in self._plugin_manager.plugins:
				return make_response("Unknown plugin: %s" % plugin_name, 404)

			plugin = self._plugin_manager.plugins[plugin_name]
			return self.command_toggle(plugin, command)

	def command_install(self, url=None, path=None, force=False, reinstall=None, dependency_links=False):
		throttled = self._get_throttled()
		if throttled and isinstance(throttled, dict) and throttled.get("current_issue", False) and not self._settings.get_boolean(["ignore_throttled"]):
			# currently throttled, we refuse to run
			return make_response("System is currently throttled, refusing to install "
			                     "anything due to possible stability issues", 409)

		if url is not None:
			if not any(map(lambda scheme: url.startswith(scheme + "://"), self.URL_SCHEMES)):
				raise ValueError("Invalid URL to pip install from")

			source = url
			source_type = "url"
			already_installed_check = lambda line: url in line

		elif path is not None:
			path = os.path.abspath(path)
			path_url = "file://" + path
			if os.sep != "/":
				# windows gets special handling
				path = path.replace(os.sep, "/").lower()
				path_url = "file:///" + path

			source = path
			source_type = "path"
			already_installed_check = lambda line: path_url in line.lower() # lower case in case of windows

		else:
			raise ValueError("Either URL or path must be provided")

		self._logger.info("Installing plugin from {}".format(source))
		pip_args = ["--disable-pip-version-check", "install", sarge.shell_quote(source), "--no-cache-dir"]

		if dependency_links or self._settings.get_boolean(["dependency_links"]):
			pip_args.append("--process-dependency-links")

		all_plugins_before = self._plugin_manager.find_plugins(existing=dict())

		try:
			returncode, stdout, stderr = self._call_pip(pip_args)

			# pip's output for a package that is already installed looks something like any of these:
			#
			#   Requirement already satisfied (use --upgrade to upgrade): OctoPrint-Plugin==1.0 from \
			#     https://example.com/foobar.zip in <lib>
			#   Requirement already satisfied (use --upgrade to upgrade): OctoPrint-Plugin in <lib>
			#   Requirement already satisfied (use --upgrade to upgrade): OctoPrint-Plugin==1.0 from \
			#     file:///tmp/foobar.zip in <lib>
			#   Requirement already satisfied (use --upgrade to upgrade): OctoPrint-Plugin==1.0 from \
			#     file:///C:/Temp/foobar.zip in <lib>
			#
			# If we detect any of these matching what we just tried to install, we'll need to trigger a second
			# install with reinstall flags.

			if not force and any(map(lambda x: x.strip().startswith(already_installed_string) and already_installed_check(x),
			                         stdout)):
				self._logger.info("Plugin to be installed from {} was already installed, forcing a reinstall".format(source))
				self._log_message("Looks like the plugin was already installed. Forcing a reinstall.")
				force = True
		except Exception:
			self._logger.exception("Could not install plugin from %s" % url)
			return make_response("Could not install plugin from URL, see the log for more details", 500)
		else:
			if force:
				# We don't use --upgrade here because that will also happily update all our dependencies - we'd rather
				# do that in a controlled manner
				pip_args += ["--ignore-installed", "--force-reinstall", "--no-deps"]
				try:
					returncode, stdout, stderr = self._call_pip(pip_args)
				except Exception:
					self._logger.exception("Could not install plugin from {}".format(source))
					return make_response("Could not install plugin from source {}, see the log for more details"
					                     .format(source), 500)

		try:
			result_line = list(filter(lambda x: x.startswith(success_string) or x.startswith(failure_string),
			                     stdout))[-1]
		except IndexError:
			self._logger.error("Installing the plugin from {} failed, could not parse output from pip. "
			                   "See plugin_pluginmanager_console.log for generated output".format(source))
			result = dict(result=False,
			              source=source,
			              source_type=source_type,
			              reason="Could not parse output from pip, see plugin_pluginmanager_console.log "
			                     "for generated output")
			self._send_result_notification("install", result)
			return jsonify(result)

		# The final output of a pip install command looks something like this:
		#
		#   Successfully installed OctoPrint-Plugin-1.0 Dependency-One-0.1 Dependency-Two-9.3
		#
		# or this:
		#
		#   Successfully installed OctoPrint-Plugin Dependency-One Dependency-Two
		#   Cleaning up...
		#
		# So we'll need to fetch the "Successfully installed" line, strip the "Successfully" part, then split
		# by whitespace and strip to get all installed packages.
		#
		# We then need to iterate over all known plugins and see if either the package name or the package name plus
		# version number matches one of our installed packages. If it does, that's our installed plugin.
		#
		# Known issue: This might return the wrong plugin if more than one plugin was installed through this
		# command (e.g. due to pulling in another plugin as dependency). It should be safe for now though to
		# consider this a rare corner case. Once it becomes a real problem we'll just extend the plugin manager
		# so that it can report on more than one installed plugin.

		result_line = result_line.strip()
		if not result_line.startswith(success_string):
			self._logger.error("Installing the plugin from {} failed, pip did not report successful installation"
			                   .format(source))
			result = dict(result=False,
			              source=source,
			              source_type=source_type,
			              reason="Pip did not report successful installation")
			self._send_result_notification("install", result)
			return jsonify(result)

		installed = list(map(lambda x: x.strip(), result_line[len(success_string):].split(" ")))
		all_plugins_after = self._plugin_manager.find_plugins(existing=dict(), ignore_uninstalled=False)

		new_plugin = self._find_installed_plugin(installed, plugins=all_plugins_after)
		if new_plugin is None:
			self._logger.warning("The plugin was installed successfully, but couldn't be found afterwards to "
			                  "initialize properly during runtime. Please restart OctoPrint.")
			result = dict(result=True,
			              source=source,
			              source_type=source_type,
			              needs_restart=True,
			              needs_refresh=True,
			              needs_reconnect=True,
			              was_reinstalled=False,
			              plugin="unknown")
			self._send_result_notification("install", result)
			return jsonify(result)

		self._plugin_manager.reload_plugins()
		needs_restart = self._plugin_manager.is_restart_needing_plugin(new_plugin) \
		                or new_plugin.key in all_plugins_before \
		                or reinstall is not None
		needs_refresh = new_plugin.implementation \
		                and isinstance(new_plugin.implementation, octoprint.plugin.ReloadNeedingPlugin)
		needs_reconnect = self._plugin_manager.has_any_of_hooks(new_plugin, self._reconnect_hooks) and self._printer.is_operational()

		is_reinstall = self._plugin_manager.is_plugin_marked(new_plugin.key, "uninstalled")
		self._plugin_manager.mark_plugin(new_plugin.key,
		                                 uninstalled=False,
		                                 installed=not is_reinstall and needs_restart)

		self._plugin_manager.log_all_plugins()

		self._logger.info("The plugin was installed successfully: {}, version {}".format(new_plugin.name, new_plugin.version))

		# noinspection PyUnresolvedReferences
		self._event_bus.fire(Events.PLUGIN_PLUGINMANAGER_INSTALL_PLUGIN, dict(id=new_plugin.key,
		                                                                      version=new_plugin.version,
		                                                                      source=source,
		                                                                      source_type=source_type))

		result = dict(result=True,
		              source=source,
		              source_type=source_type,
		              needs_restart=needs_restart,
		              needs_refresh=needs_refresh,
		              needs_reconnect=needs_reconnect,
		              was_reinstalled=new_plugin.key in all_plugins_before or reinstall is not None,
		              plugin=self._to_external_plugin(new_plugin))
		self._send_result_notification("install", result)
		return jsonify(result)

	def command_uninstall(self, plugin, cleanup=False):
		if plugin.key == "pluginmanager":
			return make_response("Can't uninstall Plugin Manager", 403)

		if not plugin.managable:
			return make_response("Plugin is not managable and hence cannot be uninstalled", 403)

		if plugin.bundled:
			return make_response("Bundled plugins cannot be uninstalled", 403)

		if plugin.origin is None:
			self._logger.warning("Trying to uninstall plugin {plugin} but origin is unknown".format(**locals()))
			return make_response("Could not uninstall plugin, its origin is unknown")

		if plugin.origin.type == "entry_point":
			# plugin is installed through entry point, need to use pip to uninstall it
			origin = plugin.origin[3]
			if origin is None:
				origin = plugin.origin[2]

			pip_args = ["--disable-pip-version-check", "uninstall", "--yes", origin]
			try:
				self._call_pip(pip_args)
			except Exception:
				self._logger.exception("Could not uninstall plugin via pip")
				return make_response("Could not uninstall plugin via pip, see the log for more details", 500)

		elif plugin.origin.type == "folder":
			import os
			import shutil
			full_path = os.path.realpath(plugin.location)

			if os.path.isdir(full_path):
				# plugin is installed via a plugin folder, need to use rmtree to get rid of it
				self._log_stdout("Deleting plugin from {folder}".format(folder=plugin.location))
				shutil.rmtree(full_path)
			elif os.path.isfile(full_path):
				self._log_stdout("Deleting plugin from {file}".format(file=plugin.location))
				os.remove(full_path)

				if full_path.endswith(".py"):
					pyc_file = "{full_path}c".format(**locals())
					if os.path.isfile(pyc_file):
						os.remove(pyc_file)

		else:
			self._logger.warning("Trying to uninstall plugin {plugin} but origin is unknown ({plugin.origin.type})".format(**locals()))
			return make_response("Could not uninstall plugin, its origin is unknown")

		needs_restart = self._plugin_manager.is_restart_needing_plugin(plugin) or cleanup
		needs_refresh = plugin.implementation and isinstance(plugin.implementation, octoprint.plugin.ReloadNeedingPlugin)
		needs_reconnect = self._plugin_manager.has_any_of_hooks(plugin, self._reconnect_hooks) and self._printer.is_operational()

		was_pending_install = self._plugin_manager.is_plugin_marked(plugin.key, "installed")
		self._plugin_manager.mark_plugin(plugin.key,
		                                 uninstalled=not was_pending_install and needs_restart,
		                                 installed=False)

		if not needs_restart:
			try:
				if plugin.enabled:
					self._plugin_manager.disable_plugin(plugin.key, plugin=plugin)
			except octoprint.plugin.core.PluginLifecycleException as e:
				self._logger.exception("Problem disabling plugin {name}".format(name=plugin.key))
				result = dict(result=False, uninstalled=True, disabled=False, unloaded=False, reason=e.reason)
				self._send_result_notification("uninstall", result)
				return jsonify(result)

			try:
				if plugin.loaded:
					self._plugin_manager.unload_plugin(plugin.key)
			except octoprint.plugin.core.PluginLifecycleException as e:
				self._logger.exception("Problem unloading plugin {name}".format(name=plugin.key))
				result = dict(result=False, uninstalled=True, disabled=True, unloaded=False, reason=e.reason)
				self._send_result_notification("uninstall", result)
				return jsonify(result)

		self._plugin_manager.reload_plugins()

		# noinspection PyUnresolvedReferences
		self._event_bus.fire(Events.PLUGIN_PLUGINMANAGER_UNINSTALL_PLUGIN, dict(id=plugin.key,
		                                                                        version=plugin.version))

		result = dict(result=True,
		              needs_restart=needs_restart,
		              needs_refresh=needs_refresh,
		              needs_reconnect=needs_reconnect,
		              plugin=self._to_external_plugin(plugin))
		self._send_result_notification("uninstall", result)
		self._logger.info("Plugin {} uninstalled".format(plugin.key))

		self._cleanup_disabled(plugin.key)
		if cleanup:
			self.command_cleanup(plugin.key, result_notifications=False)

		return jsonify(result)

	def command_cleanup(self, plugin, include_disabled=False, result_notifications=True, settings_save=True):
		if isinstance(plugin, basestring):
			key = result_value = plugin
		else:
			key = plugin.key
			result_value = self._to_external_plugin(plugin)

		message = "Cleaning up plugin {}...".format(key)
		self._logger.info(message)
		self._log_stdout(message)

		# delete plugin settings
		self._cleanup_settings(key)

		# delete plugin disabled entry
		if include_disabled:
			self._cleanup_disabled(key)

		# delete plugin data folder
		result_data = True
		if not self._cleanup_data(key):
			message = "Could not delete data folder of plugin {}".format(key)
			self._logger.exception(message)
			self._log_stderr(message)
			result_data = False

		if settings_save:
			self._settings.save()

		result = dict(result=result_data,
		              needs_restart=True,
		              plugin=result_value)
		if result_notifications:
			self._send_result_notification("cleanup", result)

		# cleaning orphan cache
		self._orphans = None

		return jsonify(result)

	def command_cleanup_all(self):
		orphans = self._get_orphans()
		cleaned_up = set()

		for orphan in sorted(orphans.keys()):
			self.command_cleanup(orphan,
			                     include_disabled=True,
			                     result_notifications=False,
			                     settings_save=False)
			cleaned_up.add(orphan)

		self._settings.save()

		result = dict(result=True,
		              needs_restart=len(cleaned_up) > 0,
		              cleaned_up=sorted(list(cleaned_up)))
		self._send_result_notification("cleanup_all", result)
		self._logger.info("Cleaned up all data, {} left overs removed".format(len(cleaned_up)))

		# cleaning orphan cache
		self._orphans = None

		return jsonify(result)

	def _cleanup_disabled(self, plugin):
		# delete from disabled list
		disabled = self._settings.global_get(["plugins", "_disabled"])
		try:
			disabled.remove(plugin)
		except ValueError:
			# not in list, ok
			pass
		self._settings.global_set(["plugins", "_disabled"], disabled)

	def _cleanup_settings(self, plugin):
		# delete plugin settings
		self._settings.global_set(["plugins", plugin], None)
		return True

	def _cleanup_data(self, plugin):
		import os
		import shutil

		data_folder = os.path.join(self._settings.getBaseFolder("data"), plugin)
		if os.path.exists(data_folder):
			try:
				shutil.rmtree(data_folder)
				return True
			except Exception:
				self._logger.exception("Could not delete plugin data folder at {}".format(data_folder))
				return False
		else:
			return True

	def command_toggle(self, plugin, command):
		if plugin.key == "pluginmanager" or (plugin.hidden and plugin.bundled):
			return make_response("Can't enable/disable Plugin Manager", 400)

		pending = ((command == "disable" and plugin.key in self._pending_enable) or (command == "enable" and plugin.key in self._pending_disable))
		safe_mode_victim = getattr(plugin, "safe_mode_victim", False)

		needs_restart = self._plugin_manager.is_restart_needing_plugin(plugin)
		needs_refresh = plugin.implementation and isinstance(plugin.implementation, octoprint.plugin.ReloadNeedingPlugin)
		needs_reconnect = self._plugin_manager.has_any_of_hooks(plugin, self._reconnect_hooks) and self._printer.is_operational()

		needs_restart_api = (needs_restart or safe_mode_victim or plugin.forced_disabled) and not pending
		needs_refresh_api = needs_refresh and not pending
		needs_reconnect_api = needs_reconnect and not pending

		try:
			if command == "disable":
				self._mark_plugin_disabled(plugin, needs_restart=needs_restart)
			elif command == "enable":
				self._mark_plugin_enabled(plugin, needs_restart=needs_restart)
		except octoprint.plugin.core.PluginLifecycleException as e:
			self._logger.exception("Problem toggling enabled state of {name}: {reason}".format(name=plugin.key, reason=e.reason))
			result = dict(result=False, reason=e.reason)
		except octoprint.plugin.core.PluginNeedsRestart:
			result = dict(result=True,
			              needs_restart=True,
			              needs_refresh=True,
			              needs_reconnect=True,
			              plugin=self._to_external_plugin(plugin))
		else:
			result = dict(result=True,
			              needs_restart=needs_restart_api,
			              needs_refresh=needs_refresh_api,
			              needs_reconnect=needs_reconnect_api,
			              plugin=self._to_external_plugin(plugin))

		self._send_result_notification(command, result)
		return jsonify(result)

	def _find_installed_plugin(self, packages, plugins=None):
		if plugins is None:
			plugins = self._plugin_manager.find_plugins(existing=dict(), ignore_uninstalled=False)

		for key, plugin in plugins.items():
			if plugin.origin is None or plugin.origin.type != "entry_point":
				continue

			package_name = plugin.origin.package_name
			package_version = plugin.origin.package_version
			versioned_package = "{package_name}-{package_version}".format(**locals())

			if package_name in packages or versioned_package in packages:
				# exact match, we are done here
				return plugin

			else:
				# it might still be a version that got stripped by python's package resources, e.g. 1.4.5a0 => 1.4.5a
				found = False

				for inst in packages:
					if inst.startswith(versioned_package):
						found = True
						break

				if found:
					return plugin

		return None

	def _send_result_notification(self, action, result):
		notification = dict(type="result", action=action)
		notification.update(result)
		self._plugin_manager.send_plugin_message(self._identifier, notification)

	def _call_pip(self, args):
		if self._pip_caller is None or not self._pip_caller.available:
			raise RuntimeError("No pip available, can't operate".format(**locals()))

		if "--process-dependency-links" in args:
			self._log_message("Installation needs to process external dependencies, that might make it take a bit longer than usual depending on the pip version")

		additional_args = self._settings.get(["pip_args"])

		if additional_args is not None:

			inapplicable_arguments = self.__class__.PIP_INAPPLICABLE_ARGUMENTS.get(args[0], list())
			for inapplicable_argument in inapplicable_arguments:
				additional_args = re.sub(r"(^|\s)" + re.escape(inapplicable_argument) + r"\\b", "", additional_args)

			if additional_args:
				args.append(additional_args)

		kwargs = dict(env=dict(PYTHONWARNINGS="ignore:DEPRECATION::pip._internal.cli.base_command"))

		return self._pip_caller.execute(*args, **kwargs)

	def _log_message(self, *lines):
		self._log(lines, prefix="*", stream="message")

	def _log_call(self, *lines):
		self._log(lines, prefix=" ", stream="call")

	def _log_stdout(self, *lines):
		self._log(lines, prefix=">", stream="stdout")

	def _log_stderr(self, *lines):
		self._log(lines, prefix="!", stream="stderr")

	def _log(self, lines, prefix=None, stream=None, strip=True):
		if strip:
			lines = list(map(lambda x: x.strip(), lines))

		self._plugin_manager.send_plugin_message(self._identifier, dict(type="loglines",
		                                                                loglines=[dict(line=line, stream=stream) for line in lines]))
		for line in lines:
			self._console_logger.debug("{prefix} {line}".format(**locals()))

	def _mark_plugin_enabled(self, plugin, needs_restart=False):
		disabled_list = list(self._settings.global_get(["plugins", "_disabled"],
		                                               validator=lambda x: isinstance(x, list),
		                                               fallback=[]))
		if plugin.key in disabled_list:
			disabled_list.remove(plugin.key)
			self._settings.global_set(["plugins", "_disabled"], disabled_list)
			self._settings.save(force=True)

		if not needs_restart and not plugin.forced_disabled and not getattr(plugin, "safe_mode_victim", False):
			self._plugin_manager.enable_plugin(plugin.key)
		else:
			if plugin.key in self._pending_disable:
				self._pending_disable.remove(plugin.key)
			elif not plugin.enabled and plugin.key not in self._pending_enable:
				self._pending_enable.add(plugin.key)

		# noinspection PyUnresolvedReferences
		self._event_bus.fire(Events.PLUGIN_PLUGINMANAGER_ENABLE_PLUGIN, dict(id=plugin.key,
		                                                                     version=plugin.version))

	def _mark_plugin_disabled(self, plugin, needs_restart=False):
		disabled_list = list(self._settings.global_get(["plugins", "_disabled"],
		                                               validator=lambda x: isinstance(x, list),
		                                               fallback=[]))
		if not plugin.key in disabled_list:
			disabled_list.append(plugin.key)
			self._settings.global_set(["plugins", "_disabled"], disabled_list)
			self._settings.save(force=True)

		if not needs_restart and not plugin.forced_disabled and not getattr(plugin, "safe_mode_victim", False):
			self._plugin_manager.disable_plugin(plugin.key)
		else:
			if plugin.key in self._pending_enable:
				self._pending_enable.remove(plugin.key)
			elif (plugin.enabled or plugin.forced_disabled or getattr(plugin, "safe_mode_victim", False)) and plugin.key not in self._pending_disable:
				self._pending_disable.add(plugin.key)

		# noinspection PyUnresolvedReferences
		self._event_bus.fire(Events.PLUGIN_PLUGINMANAGER_DISABLE_PLUGIN, dict(id=plugin.key,
		                                                                      version=plugin.version))

	def _fetch_all_data(self, do_async=False):
		def run():
			self._repository_available = self._fetch_repository_from_disk()
			self._notices_available = self._fetch_notices_from_disk()

		if do_async:
			thread = threading.Thread(target=run)
			thread.daemon = True
			thread.start()
		else:
			run()

	def _is_repository_cache_valid(self, mtime=None):
		import time
		if mtime is None:
			mtime = self._repository_mtime
		if mtime is None:
			return False
		return mtime + self._repository_cache_ttl >= time.time() > mtime

	def _fetch_repository_from_disk(self):
		repo_data = None
		if os.path.isfile(self._repository_cache_path):
			import time
			mtime = os.path.getmtime(self._repository_cache_path)
			if self._is_repository_cache_valid(mtime=mtime):
				try:
					import json
					with io.open(self._repository_cache_path, 'rt', encoding='utf-8') as f:
						repo_data = json.load(f)
					self._repository_mtime = mtime
					self._logger.info("Loaded plugin repository data from disk, was still valid")
				except Exception:
					self._logger.exception("Error while loading repository data from {}".format(self._repository_cache_path))

		return self._refresh_repository(repo_data=repo_data)

	def _fetch_repository_from_url(self):
		if not self._connectivity_checker.online:
			self._logger.info("Looks like we are offline, can't fetch repository from network")
			return None

		repository_url = self._settings.get(["repository"])
		try:
			r = requests.get(repository_url, timeout=30)
			r.raise_for_status()
			self._logger.info("Loaded plugin repository data from {}".format(repository_url))
		except Exception as e:
			self._logger.exception("Could not fetch plugins from repository at {repository_url}: {message}".format(repository_url=repository_url, message=e))
			return None

		try:
			repo_data = r.json()
		except Exception as e:
			self._logger.exception("Error while reading repository data: {}".format(e))
			return None

		# validation
		if not isinstance(repo_data, (list, tuple)):
			self._logger.warn("Invalid repository data: expected a list, got {!r}".format(repo_data))
			return None

		try:
			import json
			with octoprint.util.atomic_write(self._repository_cache_path, mode='wb') as f:
				f.write(to_bytes(json.dumps(repo_data)))
			self._repository_mtime = os.path.getmtime(self._repository_cache_path)
		except Exception as e:
			self._logger.exception("Error while saving repository data to {}: {}".format(self._repository_cache_path, e))

		return repo_data

	def _refresh_repository(self, repo_data=None):
		if repo_data is None:
			repo_data = self._fetch_repository_from_url()
			if repo_data is None:
				return False

		self._repository_plugins = list(filter(lambda x: x is not None,
		                                       map(map_repository_entry, repo_data)))
		return True

	def _is_notices_cache_valid(self, mtime=None):
		import time
		if mtime is None:
			mtime = self._notices_mtime
		if mtime is None:
			return False
		return mtime + self._notices_cache_ttl >= time.time() > mtime

	def _fetch_notices_from_disk(self):
		notice_data = None
		if os.path.isfile(self._notices_cache_path):
			import time
			mtime = os.path.getmtime(self._notices_cache_path)
			if self._is_notices_cache_valid(mtime=mtime):
				try:
					import json
					with io.open(self._notices_cache_path, 'rt', encoding='utf-8') as f:
						notice_data = json.load(f)
					self._notices_mtime = mtime
					self._logger.info("Loaded notice data from disk, was still valid")
				except Exception:
					self._logger.exception("Error while loading notices from {}".format(self._notices_cache_path))

		return self._refresh_notices(notice_data=notice_data)

	def _fetch_notices_from_url(self):
		if not self._connectivity_checker.online:
			self._logger.info("Looks like we are offline, can't fetch notices from network")
			return None

		notices_url = self._settings.get(["notices"])
		try:
			r = requests.get(notices_url, timeout=30)
			r.raise_for_status()
			self._logger.info("Loaded plugin notices data from {}".format(notices_url))
		except Exception as e:
			self._logger.exception("Could not fetch notices from {notices_url}: {message}".format(notices_url=notices_url, message=str(e)))
			return None

		notice_data = r.json()

		try:
			import json
			with octoprint.util.atomic_write(self._notices_cache_path, mode="wb") as f:
				f.write(to_bytes(json.dumps(notice_data)))
			self._notices_mtime = os.path.getmtime(self._notices_cache_path)
		except Exception as e:
			self._logger.exception("Error while saving notices to {}: {}".format(self._notices_cache_path, str(e)))
		return notice_data

	def _refresh_notices(self, notice_data=None):
		if notice_data is None:
			notice_data = self._fetch_notices_from_url()
			if notice_data is None:
				return False

		notices = dict()
		for notice in notice_data:
			if not "plugin" in notice or not "text" in notice or not "date" in notice:
				continue

			key = notice["plugin"]

			try:
				# Jekyll turns "%Y-%m-%d %H:%M:%SZ" into "%Y-%m-%d %H:%M:%S +0000", so be sure to ignore "+0000"
				#
				# Being able to use dateutil here would make things way easier but sadly that can no longer get
				# installed (from source) under OctoPi 0.14 due to its setuptools-scm dependency, so we have to do
				# without it for now until we can drop support for OctoPi 0.14.
				parsed_date = datetime.strptime(notice["date"], "%Y-%m-%d %H:%M:%S +0000")
				notice["timestamp"] = parsed_date.timetuple()
			except Exception as e:
				self._logger.warning("Error while parsing date {!r} for plugin notice "
				                  "of plugin {}, ignoring notice: {}".format(notice["date"], key,  str(e)))
				continue

			if not key in notices:
				notices[key] = []
			notices[key].append(notice)

		self._notices = notices
		return True

	def _get_orphans(self, refresh=False):
		from collections import defaultdict

		if self._orphans is not None and not refresh:
			return self._orphans

		installed_keys = self._plugin_manager.plugins.keys()
		orphans = defaultdict(lambda: dict(settings=False, data=False, disabled=False))

		# settings
		for key in list(self._settings.global_get(["plugins"]).keys()):
			if key.startswith("_"):
				# internal key, like _disabled
				continue

			if key not in installed_keys:
				orphans[key]["settings"] = True

		# data
		for entry in scandir(self._settings.getBaseFolder("data")):
			if not entry.is_dir():
				continue

			if entry.name not in installed_keys:
				orphans[entry.name]["data"] = True

		# disabled
		disabled = self._settings.global_get(["plugins", "_disabled"])
		for key in disabled:
			if key not in installed_keys:
				orphans[key]["disabled"] = True

		self._orphans = dict(**orphans)
		return self._orphans

	@property
	def _reconnect_hooks(self):
		reconnect_hooks = self.__class__.RECONNECT_HOOKS

		reconnect_hook_provider_hooks = self._plugin_manager.get_hooks("octoprint.plugin.pluginmanager.reconnect_hooks")
		for name, hook in reconnect_hook_provider_hooks.items():
			try:
				result = hook()
				if isinstance(result, (list, tuple)):
					reconnect_hooks.extend(filter(lambda x: isinstance(x, basestring), result))
			except Exception:
				self._logger.exception("Error while retrieving additional hooks for which a "
				                       "reconnect is required from plugin {name}".format(**locals()),
				                       extra=dict(plugin=name))

		return reconnect_hooks

	def _get_plugins(self):
		plugins = self._plugin_manager.plugins

		hidden = self._settings.get(["hidden"])
		result = []
		for key, plugin in plugins.items():
			if key in hidden or (plugin.bundled and plugin.hidden):
				continue
			result.append(self._to_external_plugin(plugin))

		return result

	def _to_external_plugin(self, plugin):
		return dict(
			key=plugin.key,
			name=plugin.name,
			description=plugin.description,
			disabling_discouraged=gettext(plugin.disabling_discouraged) if plugin.disabling_discouraged else False,
			author=plugin.author,
			version=plugin.version,
			url=plugin.url,
			license=plugin.license,
			bundled=plugin.bundled,
			managable=plugin.managable,
			enabled=plugin.enabled,
			blacklisted=plugin.blacklisted,
			forced_disabled=plugin.forced_disabled,
			incompatible=plugin.incompatible,
			safe_mode_victim=getattr(plugin, "safe_mode_victim", False),
			pending_enable=(not plugin.enabled and not getattr(plugin, "safe_mode_victim", False) and plugin.key in self._pending_enable),
			pending_disable=((plugin.enabled or getattr(plugin, "safe_mode_victim", False)) and plugin.key in self._pending_disable),
			pending_install=(self._plugin_manager.is_plugin_marked(plugin.key, "installed")),
			pending_uninstall=(self._plugin_manager.is_plugin_marked(plugin.key, "uninstalled")),
			origin=plugin.origin.type,
			notifications = self._get_notifications(plugin)
		)

	def _get_notifications(self, plugin):
		key = plugin.key
		if not plugin.enabled:
			return

		if key not in self._notices:
			return

		octoprint_version = get_octoprint_version(base=True)
		plugin_notifications = self._notices.get(key, [])

		def filter_relevant(notification):
			return "text" in notification and "date" in notification and \
			       ("versions" not in notification or plugin.version in notification["versions"]) and \
			       ("octoversions" not in notification or is_octoprint_compatible(*notification["octoversions"],
			                                                                      octoprint_version=octoprint_version))

		def map_notification(notification):
			return self._to_external_notification(key, notification)

		return list(filter(lambda x: x is not None,
		              	   map(map_notification,
		                  	   filter(filter_relevant,
		                         	  plugin_notifications))))

	def _to_external_notification(self, key, notification):
		return dict(key=key,
		            date=time.mktime(notification["timestamp"]),
		            text=notification["text"],
		            link=notification.get("link"),
		            versions=notification.get("versions", []),
		            important=notification.get("important", False))


def _register_custom_events(*args, **kwargs):
	return ["install_plugin", "uninstall_plugin", "enable_plugin", "disable_plugin"]


__plugin_name__ = "Plugin Manager"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "http://docs.octoprint.org/en/master/bundledplugins/pluginmanager.html"
__plugin_description__ = "Allows installing and managing OctoPrint plugins"
__plugin_license__ = "AGPLv3"
__plugin_hidden__ = True

def __plugin_load__():
	global __plugin_implementation__
	__plugin_implementation__ = PluginManagerPlugin()

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.server.http.bodysize": __plugin_implementation__.increase_upload_bodysize,
		"octoprint.ui.web.templatetypes": __plugin_implementation__.get_template_types,
		"octoprint.events.register_custom_events": _register_custom_events,
		"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
	}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import requests
import time


def build_base_url(https=False, httpuser=None, httppass=None, host=None, port=None, prefix=None):
	protocol = "https" if https else "http"
	httpauth = "{}:{}@".format(httpuser, httppass) if httpuser and httppass else ""
	host = host if host else "127.0.0.1"
	port = ":{}".format(port) if port else ":5000"
	prefix = prefix if prefix else ""

	return "{}://{}{}{}{}".format(protocol, httpauth, host, port, prefix)


class SocketTimeout(BaseException):
	pass

class SocketClient(object):
	def __init__(self, url, use_ssl=False, daemon=True, **kwargs):
		self._url = url
		self._use_ssl = use_ssl
		self._daemon = daemon
		self._ws_kwargs = kwargs

		self._seen_open = False
		self._seen_close = False
		self._waiting_for_reconnect = False

		self._ws = None
		self._thread = None

		# hello world

	def _prepare(self):
		"""Prepares socket and thread for a new connection."""

		# close the socket if it's currently open
		if self._ws is not None:
			try:
				self._ws.close()
			except Exception:
				# we can't handle that in any meaningful way right now
				pass

		# prepare a bunch of callback methods
		import functools
		callbacks = dict()
		for callback in ("on_open", "on_message", "on_error", "on_close"):
			# now, normally we could just use functools.partial for something like
			# this, but websocket does a type check against a python function type
			# which the use of partial makes fail, so we have to use a lambda
			# wrapper here, and since we need to pick the current value of
			# callback from the current scope we need a factory method for that...
			def factory(cb):
				return lambda *fargs, **fkwargs: functools.partial(self._on_callback, cb)(*fargs, **fkwargs)
			callbacks[callback] = factory(callback)

		# initialize socket instance with url and callbacks
		import websocket
		kwargs = dict(self._ws_kwargs)
		kwargs.update(callbacks)
		self._ws = websocket.WebSocketApp(self._url, **kwargs)

		# initialize thread
		import threading
		self._thread = threading.Thread(target=self._on_thread_run)
		self._thread.daemon = self._daemon

	def _on_thread_run(self):
		"""Has the socket run forever (aka until closed...)."""
		self._ws.run_forever()

	def _on_callback(self, cb, *args, **kwargs):
		"""
		Callback for socket events.

		Will call any callback method defined on ``self`` that matches ``cb``
		prefixed with an ``_ws_``, then will call any callback method provided in
		the socket keyword arguments (``self._ws_kwargs``) that matches ``cb``.

		Calling args and kwargs will be the ones passed to ``_on_callback``.

		Arguments:
		    cb (str): the callback type
		"""
		internal = "_ws_" + cb
		if hasattr(self, internal):
			cb_func = getattr(self, internal)
			if callable(cb_func):
				cb_func(*args, **kwargs)

		cb_func = self._ws_kwargs.get(cb, None)
		if callable(cb_func):
			cb_func(*args, **kwargs)

	def _ws_on_open(self, ws):
		"""
		Callback for socket on_open event.

		Used only to track active reconnection attempts.
		"""

		if not self._waiting_for_reconnect:
			return
		if ws != self._ws:
			return
		self._seen_open = True

	def _ws_on_close(self, ws):
		"""
		Callback for socket on_close event.

		Used only to track active reconnection attempts.
		"""

		if not self._waiting_for_reconnect:
			return
		if ws != self._ws:
			return
		self._seen_close = True

	def connect(self):
		"""Connects the socket."""
		self._prepare()
		self._thread.start()

	def wait(self, timeout=None):
		"""Waits for the closing of the socket or the timeout."""
		start = None

		def test_condition():
			if timeout and start and start + timeout > time.time():
				raise SocketTimeout()

		start = time.time()
		while self._thread.is_alive():
			test_condition()
			self._thread.join(timeout=1.0)

	@property
	def is_connected(self):
		"""Whether the web socket is connected or not."""
		return self._thread and self._ws and self._thread.is_alive()

	def disconnect(self):
		"""Disconnect the web socket."""
		if self._ws:
			self._ws.close()

	def reconnect(self, timeout=None, disconnect=True):
		"""
		Tries to reconnect to the web socket.

		If timeout is set will try to reconnect over the specified timeout in seconds
		and return False if the connection could not be re-established.

		If no timeout is provided, the method will block until the connection could
		be re-established.

		If disconnect is set to ``True`` will disconnect the socket explicitly
		first if it is currently connected.

		Arguments:
		    timeout (number): timeout in seconds to wait for the reconnect to happen.
		    disconnect (bool): Whether to disconnect explicitly from the socket if
		       a connection is currently established (True, default) or not (False).

		Returns:
		    bool - True if the reconnect was successful, False otherwise.
		"""

		self._seen_close = False
		self._seen_open = False
		self._waiting_for_reconnect = True

		if not self.is_connected:
			# not connected, so we are already closed
			self._seen_close = True
		elif disconnect:
			# connected and disconnect is True, so we disconnect
			self.disconnect()

		start = None
		if timeout:
			timeout_condition = lambda: start is not None and time.time() > start + timeout
		else:
			timeout_condition = lambda: False

		start = time.time()
		while not timeout_condition():
			if self._seen_close and self._seen_open:
				# we saw a connection close and open, so a reconnect, success!
				return True
			else:
				# try to connect
				self.connect()

			# sleep a bit
			time.sleep(1.0)

		# if we land here the timeout condition became True without us seeing
		# a reconnect, that's a failure
		return False

class Client(object):

	def __init__(self, baseurl, apikey):
		self.baseurl = baseurl
		self.apikey = apikey

	def prepare_request(self, method=None, path=None, params=None):
		url = None
		if self.baseurl:
			while path.startswith("/"):
				path = path[1:]
			url = self.baseurl + "/" + path
		return requests.Request(method=method, url=url, params=params, headers={"X-Api-Key": self.apikey}).prepare()

	def request(self, method, path, data=None, files=None, encoding=None, params=None, timeout=None):
		if timeout is None:
			timeout = 30

		s = requests.Session()
		request = self.prepare_request(method, path, params=params)
		if data or files:
			if encoding == "json":
				request.prepare_body(None, None, json=data)
			else:
				request.prepare_body(data, files=files)
		response = s.send(request, timeout=timeout)
		return response

	def get(self, path, params=None, timeout=None):
		return self.request("GET", path, params=params, timeout=timeout)

	def post(self, path, data, encoding=None, params=None, timeout=None):
		return self.request("POST", path, data=data, encoding=encoding, params=params, timeout=timeout)

	def post_json(self, path, data, params=None, timeout=None):
		return self.post(path, data, encoding="json", params=params, timeout=timeout)

	def post_command(self, path, command, additional=None, timeout=None):
		data = dict(command=command)
		if additional:
			data.update(additional)
		return self.post_json(path, data, params=data, timeout=timeout)

	def upload(self, path, file_path, additional=None, file_name=None, content_type=None, params=None, timeout=None):
		import os

		if not os.path.isfile(file_path):
			raise ValueError("{} cannot be uploaded since it is not a file".format(file_path))

		if file_name is None:
			file_name = os.path.basename(file_path)

		with io.open(file_path, 'rb') as fp:
			if content_type:
				files = dict(file=(file_name, fp, content_type))
			else:
				files = dict(file=(file_name, fp))

			response = self.request("POST", path, data=additional, files=files, params=params, timeout=timeout)

		return response

	def delete(self, path, params=None, timeout=None):
		return self.request("DELETE", path, params=params, timeout=timeout)

	def patch(self, path, data, encoding=None, params=None, timeout=None):
		return self.request("PATCH", path, data=data, encoding=encoding, params=params, timeout=timeout)

	def put(self, path, data, encoding=None, params=None, timeout=None):
		return self.request("PUT", path, data=data, encoding=encoding, params=params, timeout=timeout)

	def create_socket(self, **kwargs):
		import uuid
		import random
		import json

		# creates websocket URL for SockJS according to
		# - http://sockjs.github.io/sockjs-protocol/sockjs-protocol-0.3.3.html#section-37
		# - http://sockjs.github.io/sockjs-protocol/sockjs-protocol-0.3.3.html#section-50
		url = "ws://{}/sockjs/{:0>3d}/{}/websocket".format(
			self.baseurl[self.baseurl.find("//") + 2:], # host + port + prefix, but no protocol
			random.randrange(0, stop=999),              # server_id
			uuid.uuid4()                                # session_id
		)
		use_ssl = self.baseurl.startswith("https:")

		on_open_cb = kwargs.get("on_open", None)
		on_heartbeat_cb = kwargs.get("on_heartbeat", None)
		on_message_cb = kwargs.get("on_message", None)
		on_close_cb = kwargs.get("on_close", None)
		on_error_cb = kwargs.get("on_error", None)
		daemon = kwargs.get("daemon", True)

		def on_message(ws, message):
			message_type = message[0]

			if message_type == "h":
				# "heartbeat" message
				if callable(on_heartbeat_cb):
					on_heartbeat_cb(ws)
					return
			elif message_type == "o":
				# "open" message
				return
			elif message_type == "c":
				# "close" message
				return

			if not callable(on_message_cb):
				return

			message_body = message[1:]
			if not message_body:
				return

			data = json.loads(message_body)

			if message_type == "m":
				data = [data,]

			for d in data:
				for internal_type, internal_message in d.items():
					on_message_cb(ws, internal_type, internal_message)

		def on_open(ws):
			if callable(on_open_cb):
				on_open_cb(ws)

		def on_close(ws):
			if callable(on_close_cb):
				on_close_cb(ws)

		def on_error(ws, error):
			if callable(on_error_cb):
				on_error_cb(ws, error)

		socket = SocketClient(url,
		                      use_ssl=use_ssl,
		                      daemon=daemon,
		                      on_open=on_open,
		                      on_message=on_message,
		                      on_close=on_close,
		                      on_error=on_error)
		socket.connect()

		return socket



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from flask_babel import gettext

from octoprint.access.permissions import Permissions

from collections import defaultdict
import threading

class LoginUiPlugin(octoprint.plugin.UiPlugin,
                    octoprint.plugin.TemplatePlugin,
                    octoprint.plugin.AssetPlugin):
	MAX_BACKLOG_LEN = 100

	# noinspection PyMissingConstructor
	def __init__(self):
		self._message_backlog = defaultdict(list)
		self._message_backlog_mutex = threading.RLock()

	@property
	def active(self):
		# we are only active if ACL is enabled AND configured
		return self._user_manager.enabled and self._user_manager.has_been_customized()

	def get_assets(self):
		return dict(
			js=["js/viewmodel.js"]
		)

	def will_handle_ui(self, request):
		if not self.active:
			# not active, not responsible
			return False

		from octoprint.server.util import loginUserFromApiKey, loginUserFromAuthorizationHeader, InvalidApiKeyException
		from octoprint.server.util.flask import passive_login

		# first try to login via api key & authorization header, just in case that's set
		try:
			if loginUserFromApiKey():
				# successful? No need for handling the UI
				return False
		except InvalidApiKeyException:
			pass # ignored

		if loginUserFromAuthorizationHeader():
			# successful? No need for handling the UI
			return False

		# then try a passive login
		passive_login()
		if Permissions.STATUS.can() and Permissions.SETTINGS_READ.can():
			# Status & settings_read permission? No need to handle UI
			return False
		else:
			return True

	def on_ui_render(self, now, request, render_kwargs):
		from flask import render_template, make_response

		def add_additional_assets(hook):
			result = []
			for name, hook in self._plugin_manager.get_hooks(hook).items():
				try:
					assets = hook()
					if isinstance(assets, (tuple, list)):
						result += assets
				except:
					self._logger.exception("Error fetching theming CSS to include from plugin {}".format(name),
					                       extra=dict(plugin=name))
			return result

		additional_assets = []
		additional_assets += add_additional_assets("octoprint.plugin.loginui.theming")

		# backwards compatibility to old forcelogin plugin which was replaced by this one
		additional_assets += add_additional_assets("octoprint.plugin.forcelogin.theming")

		render_kwargs.update(dict(loginui_theming=additional_assets))
		return make_response(render_template("loginui_index.jinja2", **render_kwargs))

	def get_ui_custom_tracked_files(self):
		from os.path import join as opj

		paths = [opj("static", "css", "loginui.css"),
		         opj("static", "js", "main.js"),
		         opj("static", "js", "viewmodel.js"),
		         opj("static", "less", "loginui.less"),
		         opj("templates", "parts", "loginui_css.jinja2"),
		         opj("templates", "parts", "loginui_javascripts.jinja2"),
		         opj("templates", "loginui_index.jinja2")]

		return [opj(self._basefolder, path) for path in paths]

	def get_ui_preemptive_caching_enabled(self):
		return False

	def get_sorting_key(self, context=None):
		if context == "UiPlugin.on_ui_render":
			# If a plugin *really* wants to come before this plugin, it'll have to turn to negative numbers.
			#
			# This is obviously discouraged for security reasons, but very specific setups might make it necessary,
			# so we make it possible. If this should get abused long term we can always turn this into -inf.
			return 0

__plugin_name__ = "Login UI"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Displays a standalone login UI if needed"
__plugin_disabling_discouraged__ = gettext("Without this plugin there will be no dedicated login page for users in "
                                           "case anonymous read only access is disabled - instead your non logged in "
                                           "visitors will get a broken UI.")
__plugin_hidden__ = True

__plugin_implementation__ = LoginUiPlugin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.router
    ~~~~~~~~~~~~~~~~~~~~~

    SockJS protocol router implementation.
"""

from tornado import ioloop, version_info

from octoprint.vendor.sockjs.tornado import transports, session, sessioncontainer, static, stats, proto


DEFAULT_SETTINGS = {
    # Sessions check interval in seconds
    'session_check_interval': 1,
    # Session expiration in seconds
    'disconnect_delay': 5,
    # Heartbeat time in seconds. Do not change this value unless
    # you absolutely sure that new value will work.
    'heartbeat_delay': 25,
    # Enabled protocols
    'disabled_transports': [],
    # SockJS location
    'sockjs_url': 'https://cdn.jsdelivr.net/sockjs/0.3/sockjs.min.js',
    # Max response body size
    'response_limit': 128 * 1024,
    # Enable or disable JSESSIONID cookie handling
    'jsessionid': True,
    # Should sockjs-tornado flush messages immediately or queue then and
    # flush on next ioloop tick
    'immediate_flush': True,
    # Enable or disable Nagle for persistent transports
    'disable_nagle': True,
    # Enable IP checks for polling transports. If enabled, all subsequent
    # polling calls should be from the same IP address.
    'verify_ip': True,
    # list of allowed origins for websocket connections
    # or "*" - accept all websocket connections
    'websocket_allow_origin': "*"
    }

GLOBAL_HANDLERS = [
    ('xhr_send', transports.XhrSendHandler),
    ('jsonp_send', transports.JSONPSendHandler)
]

TRANSPORTS = {
    'websocket': transports.WebSocketTransport,
    'xhr': transports.XhrPollingTransport,
    'xhr_streaming': transports.XhrStreamingTransport,
    'jsonp': transports.JSONPTransport,
    'eventsource': transports.EventSourceTransport,
    'htmlfile': transports.HtmlFileTransport
}

STATIC_HANDLERS = {
    '/chunking_test': static.ChunkingTestHandler,
    '/info': static.InfoHandler,
    '/iframe[0-9-.a-z_]*.html': static.IFrameHandler,
    '/websocket': transports.RawWebSocketTransport,
    '/?': static.GreetingsHandler
}


class SockJSRouter(object):
    """SockJS protocol router"""
    def __init__(self,
                 connection,
                 prefix='',
                 user_settings=dict(),
                 io_loop=None,
                 session_kls=None):
        """Constructor.

        `connection`
            SockJSConnection class
        `prefix`
            Connection prefix
        `user_settings`
            Settings dictionary
        `io_loop`
            Optional IOLoop instance
        """

        # TODO: Version check
        if version_info[0] < 2:
            raise Exception('sockjs-tornado requires Tornado 2.0 or higher.')

        # Store connection class
        self._connection = connection

        # Initialize io_loop
        self.io_loop = io_loop or ioloop.IOLoop.instance()

        # Settings
        self.settings = DEFAULT_SETTINGS.copy()
        if user_settings:
            self.settings.update(user_settings)

        self.websockets_enabled = 'websocket' not in self.settings['disabled_transports']
        self.cookie_needed = self.settings['jsessionid']

        # Sessions
        self._session_kls = session_kls if session_kls else session.Session
        self._sessions = sessioncontainer.SessionContainer()

        check_interval = self.settings['session_check_interval'] * 1000
        self._sessions_cleanup = ioloop.PeriodicCallback(self._sessions.expire,
                                                         check_interval,
                                                         self.io_loop)
        self._sessions_cleanup.start()

        # Stats
        self.stats = stats.StatsCollector(self.io_loop)

        # Initialize URLs
        base = prefix + r'/[^/.]+/(?P<session_id>[^/.]+)'

        # Generate global handler URLs
        self._transport_urls = [('%s/%s$' % (base, p[0]), p[1], dict(server=self))
                                for p in GLOBAL_HANDLERS]

        for k, v in TRANSPORTS.items():
            if k in self.settings['disabled_transports']:
                continue

            # Only version 1 is supported
            self._transport_urls.append(
                (r'%s/%s$' % (base, k),
                 v,
                 dict(server=self))
                )

        # Generate static URLs
        self._transport_urls.extend([('%s%s' % (prefix, k), v, dict(server=self))
                                     for k, v in STATIC_HANDLERS.items()])

    @property
    def urls(self):
        """List of the URLs to be added to the Tornado application"""
        return self._transport_urls

    def apply_routes(self, routes):
        """Feed list of the URLs to the routes list. Returns list"""
        routes.extend(self._transport_urls)
        return routes

    def create_session(self, session_id, register=True):
        """Creates new session object and returns it.

        `request`
            Request that created the session. Will be used to get query string
            parameters and cookies
        `register`
            Should be session registered in a storage. Websockets don't
            need it.
        """
        # TODO: Possible optimization here for settings.get
        s = self._session_kls(self._connection,
                            self,
                            session_id,
                            self.settings.get('disconnect_delay')
                            )

        if register:
            self._sessions.add(s)

        return s

    def get_session(self, session_id):
        """Get session by session id

        `session_id`
            Session id
        """
        return self._sessions.get(session_id)

    def get_connection_class(self):
        """Return associated connection class"""
        return self._connection

    # Broadcast helper
    def broadcast(self, clients, msg):
        """Optimized `broadcast` implementation. Depending on type of the session, will json-encode
        message once and will call either `send_message` or `send_jsonifed`.

        `clients`
            Clients iterable
        `msg`
            Message to send
        """
        json_msg = None

        count = 0

        for c in clients:
            sess = c.session
            if not sess.is_closed:
                if sess.send_expects_json:
                    if json_msg is None:
                        json_msg = proto.json_encode(msg)
                    sess.send_jsonified(json_msg, False)
                else:
                    sess.send_message(msg, stats=False)

                count += 1

        self.stats.on_pack_sent(count)
